from collections.abc import Hashable, Mapping, Sequence, Set
from typing import Any, NamedTuple, Union, overload

from pydantic import BaseModel


def hash_collection(
    collection: Union[Mapping[Hashable, Any], Sequence[Any], Set[Any], tuple[Any, ...], NamedTuple],
) -> int:
    """Hash a mutable collection or immutable collection containing mutable elements.

    This is useful for hashing Dagster-specific NamedTuples that contain mutable lists or dicts.
    The default NamedTuple __hash__ function assumes the contents of the NamedTuple are themselves
    hashable, and will throw an error if they are not. This can occur when trying to e.g. compute a
    cache key for the tuple for use with `lru_cache`.

    This alternative implementation will recursively process collection elements to convert basic
    lists and dicts to tuples prior to hashing. It is recommended to cache the result:

    Example:
        .. code-block:: python

            def __hash__(self):
                if not hasattr(self, '_hash'):
                    self._hash = hash_named_tuple(self)
                return self._hash
    """
    assert isinstance(collection, (list, dict, set, tuple)), (
        f"Cannot hash collection of type {type(collection)}"
    )
    return hash(make_hashable(collection))


@overload
def make_hashable(value: Union[list[Any], set[Any]]) -> tuple[Any, ...]: ...


@overload
def make_hashable(value: dict[Any, Any]) -> tuple[tuple[Any, Any]]: ...


@overload
def make_hashable(value: Any) -> Any: ...


def make_hashable(value: Any) -> Any:
    from dagster_shared.record import as_dict, is_record

    if isinstance(value, dict):
        return tuple(sorted((key, make_hashable(val)) for key, val in value.items()))
    elif is_record(value):
        return tuple(make_hashable(val) for val in as_dict(value).values())
    elif isinstance(value, (list, tuple, set)):
        return tuple([make_hashable(x) for x in value])
    elif isinstance(value, BaseModel):
        return make_hashable(value.dict())
    else:
        return value



from functools import cached_property
from typing import TYPE_CHECKING, Any, Optional

from pydantic import BaseModel, ConfigDict
from typing_extensions import Self


class DagsterModel(BaseModel):
    """Standardizes on Pydantic settings that are stricter than the default.
    - Frozen, to avoid complexity caused by mutation.
    - extra=forbid, to avoid bugs caused by accidentally constructing with the wrong arguments.
    - arbitrary_types_allowed, to allow non-model class params to be validated with isinstance.
    - Avoid pydantic reading a cached property class as part of the schema.
    """

    if TYPE_CHECKING:
        # without this, the type checker does not understand the constructor kwargs on subclasses
        def __init__(self, **data: Any) -> None: ...

    model_config = ConfigDict(
        extra="forbid",
        frozen=True,
        arbitrary_types_allowed=True,
        ignored_types=(cached_property,),
    )

    def model_copy(self, *, update: Optional[dict[str, Any]] = None) -> Self:  # pyright: ignore[reportIncompatibleMethodOverride]
        return super().model_copy(update=update)

    @classmethod
    def model_construct(cls, **kwargs: Any) -> Self:  # pyright: ignore[reportIncompatibleMethodOverride]
        return super().model_construct(**kwargs)



import datetime
import hashlib
from typing import NamedTuple, Optional

from dagster_shared.serdes.serdes import (
    PackableValue,
    WhitelistMap,
    serialize_value,
    whitelist_for_serdes,
)


def create_snapshot_id(
    snapshot: PackableValue, whitelist_map: Optional[WhitelistMap] = None
) -> str:
    kwargs = dict(whitelist_map=whitelist_map) if whitelist_map else {}
    json_rep = serialize_value(snapshot, **kwargs)
    return hash_str(json_rep)


def hash_str(in_str: str) -> str:
    # so that hexdigest is 40, not 64 bytes
    return hashlib.sha1(in_str.encode("utf-8")).hexdigest()


def serialize_pp(value: PackableValue) -> str:
    """Serialize and pretty print."""
    return serialize_value(value, indent=2, separators=(",", ": "))


@whitelist_for_serdes
class SerializableTimeDelta(NamedTuple):
    """A Dagster-serializable version of a datetime.timedelta. The datetime.timedelta class
    internally stores values as an integer number of days, seconds, and microseconds. This class
    handles converting between the in-memory and serializable formats.

    This class should not be used directly in application code. Any place it pops up should be converted to `datetime.timedelta`
    using `to_timedelta()`.

    Consequently, also do not rely on `days`, `seconds`, or `microseconds` attributes for any datetime arithmetic.
    """

    days: int
    seconds: int
    microseconds: int

    @staticmethod
    def from_timedelta(timedelta: datetime.timedelta) -> "SerializableTimeDelta":
        return SerializableTimeDelta(
            days=timedelta.days, seconds=timedelta.seconds, microseconds=timedelta.microseconds
        )

    def to_timedelta(self) -> datetime.timedelta:
        return datetime.timedelta(
            days=self.days, seconds=self.seconds, microseconds=self.microseconds
        )



"""Serialization & deserialization for Dagster objects.

Why have custom serialization?

* Default json serialization doesn't work well on namedtuples, which we use extensively to create
  immutable value types. Namedtuples serialize like tuples as flat lists.
* Explicit whitelisting should help ensure we are only persisting or communicating across a
  serialization boundary the types we expect to.

Why not pickle?

* This isn't meant to replace pickle in the conditions that pickle is reasonable to use
  (in memory, not human readable, etc) just handle the json case effectively.
"""

import collections.abc
import dataclasses
from abc import ABC, abstractmethod
from collections.abc import Iterable, Iterator, Mapping, Sequence
from dataclasses import is_dataclass
from enum import Enum
from functools import cached_property, partial
from inspect import Parameter, signature
from typing import (  # noqa: UP035
    TYPE_CHECKING,
    AbstractSet,
    Any,
    Callable,
    Final,
    Generic,
    NamedTuple,
    Optional,
    Union,
    cast,
    overload,
)

from typing_extensions import Self, TypeAlias, TypeVar

import dagster_shared.check as check
from dagster_shared import seven
from dagster_shared.match import is_named_tuple_subclass, match_type
from dagster_shared.record import (
    IHaveNew,
    as_dict_for_new,
    get_record_annotations,
    has_generated_new,
    is_record,
)
from dagster_shared.serdes.errors import DeserializationError, SerdesUsageError, SerializationError
from dagster_shared.utils.warnings import disable_dagster_warnings

if TYPE_CHECKING:
    # There is no actual class backing Dataclasses, _typeshed provides this
    # protocol.
    from _typeshed import DataclassInstance
    from pydantic import BaseModel


###################################################################################################
# Types
###################################################################################################


JsonSerializableValue: TypeAlias = Union[
    Sequence["JsonSerializableValue"],
    Mapping[str, "JsonSerializableValue"],
    str,
    int,
    float,
    bool,
    None,
]

PackableValue: TypeAlias = Union[
    Sequence["PackableValue"],
    Mapping[str, "PackableValue"],
    str,
    int,
    float,
    bool,
    None,
    NamedTuple,
    "BaseModel",
    "DataclassInstance",
    set["PackableValue"],
    frozenset["PackableValue"],
    Enum,
    IHaveNew,  # indirect way of indicating @record_custom classes are packable
]

UnpackedValue: TypeAlias = Union[
    Sequence["UnpackedValue"],
    Mapping[str, "UnpackedValue"],
    str,
    int,
    float,
    bool,
    None,
    NamedTuple,
    "BaseModel",
    "DataclassInstance",
    set["PackableValue"],
    frozenset["PackableValue"],
    Enum,
    "UnknownSerdesValue",
    IHaveNew,
]

SerializableObject: TypeAlias = Union[
    NamedTuple,
    "BaseModel",
    "DataclassInstance",
]

_K = TypeVar("_K")
_V = TypeVar("_V")


class SerializableNonScalarKeyMapping(Mapping[_K, _V]):
    """Wrapper class for non-scalar key mappings, used to performantly type check when serializing
    without impacting the performance of serializing the more common scalar key dicts.
    May be replaceable with a different clever scheme.
    """

    def __init__(self, mapping: Mapping[_K, _V] = {}) -> None:
        self.mapping: Mapping[_K, _V] = mapping

    def __setitem__(self, key: _K, item: _V):
        raise NotImplementedError("SerializableNonScalarKeyMapping is immutable")

    def __getitem__(self, item: _K) -> _V:
        return self.mapping[item]

    def __len__(self) -> int:
        return len(self.mapping)

    def __iter__(self) -> Iterator[_K]:
        return iter(self.mapping)


###################################################################################################
# Whitelisting
###################################################################################################


class WhitelistMap(NamedTuple):
    object_serializers: dict[str, "ObjectSerializer"]
    object_deserializers: dict[str, "ObjectSerializer"]
    enum_serializers: dict[str, "EnumSerializer"]
    object_type_map: dict[str, type]

    def register_object(
        self,
        name: str,
        object_class: type,
        serializer_class: type["ObjectSerializer"],
        storage_name: Optional[str] = None,
        old_storage_names: Optional[AbstractSet[str]] = None,
        storage_field_names: Optional[Mapping[str, str]] = None,
        old_fields: Optional[Mapping[str, JsonSerializableValue]] = None,
        skip_when_empty_fields: Optional[AbstractSet[str]] = None,
        skip_when_none_fields: Optional[AbstractSet[str]] = None,
        field_serializers: Optional[Mapping[str, type["FieldSerializer"]]] = None,
        kwargs_fields: Optional[AbstractSet[str]] = None,
    ):
        """Register a model class in the whitelist map.

        Args:
            name: The class name of the namedtuple to register
            object_class: The object class to register.
                Can be None to gracefull load previously serialized objects as None.
            serializer: The class to use when serializing and deserializing
        """
        serializer = serializer_class(
            klass=object_class,
            storage_name=storage_name,
            storage_field_names=storage_field_names,
            old_fields=old_fields,
            skip_when_empty_fields=skip_when_empty_fields,
            skip_when_none_fields=skip_when_none_fields,
            field_serializers=(
                {k: klass.get_instance() for k, klass in field_serializers.items()}
                if field_serializers
                else None
            ),
            kwargs_fields=kwargs_fields,
        )
        self.object_serializers[name] = serializer
        deserializer_name = storage_name or name
        if deserializer_name in self.object_deserializers:
            raise SerdesUsageError(
                f"Multiple deserializers registered for storage name `{deserializer_name}`"
            )
        self.object_deserializers[deserializer_name] = serializer
        if old_storage_names:
            for old_storage_name in old_storage_names:
                self.object_deserializers[old_storage_name] = serializer

        self.object_type_map[name] = serializer.klass

    def register_enum(
        self,
        name: str,
        enum_class: type[Enum],
        serializer_class: Optional[type["EnumSerializer"]] = None,
        storage_name: Optional[str] = None,
        old_storage_names: Optional[AbstractSet[str]] = None,
    ) -> None:
        serializer_class = serializer_class or EnumSerializer
        serializer = serializer_class(
            klass=enum_class,
            storage_name=storage_name,
        )
        self.enum_serializers[name] = serializer
        if storage_name:
            self.enum_serializers[storage_name] = serializer
        if old_storage_names:
            for old_storage_name in old_storage_names:
                self.enum_serializers[old_storage_name] = serializer

    @staticmethod
    def create() -> "WhitelistMap":
        return WhitelistMap(
            object_serializers={},
            object_deserializers={},
            enum_serializers={},
            object_type_map={},
        )


_WHITELIST_MAP: Final[WhitelistMap] = WhitelistMap.create()

T = TypeVar("T")
U = TypeVar("U")
T_Type = TypeVar("T_Type", bound=type[object])
T_Scalar = TypeVar("T_Scalar", bound=Union[str, int, float, bool, None])


@overload
def whitelist_for_serdes(__cls: T_Type) -> T_Type: ...


@overload
def whitelist_for_serdes(
    __cls: None = None,
    *,
    serializer: Optional[type["Serializer"]] = ...,
    storage_name: Optional[str] = ...,
    old_storage_names: Optional[AbstractSet[str]] = None,
    storage_field_names: Optional[Mapping[str, str]] = ...,
    old_fields: Optional[Mapping[str, JsonSerializableValue]] = ...,
    skip_when_empty_fields: Optional[AbstractSet[str]] = ...,
    skip_when_none_fields: Optional[AbstractSet[str]] = ...,
    field_serializers: Optional[Mapping[str, type["FieldSerializer"]]] = None,
    kwargs_fields: Optional[AbstractSet[str]] = None,
) -> Callable[[T_Type], T_Type]: ...


def whitelist_for_serdes(
    __cls: Optional[T_Type] = None,
    *,
    serializer: Optional[type["Serializer"]] = None,
    storage_name: Optional[str] = None,
    old_storage_names: Optional[AbstractSet[str]] = None,
    storage_field_names: Optional[Mapping[str, str]] = None,
    old_fields: Optional[Mapping[str, JsonSerializableValue]] = None,
    skip_when_empty_fields: Optional[AbstractSet[str]] = None,
    skip_when_none_fields: Optional[AbstractSet[str]] = None,
    field_serializers: Optional[Mapping[str, type["FieldSerializer"]]] = None,
    kwargs_fields: Optional[AbstractSet[str]] = None,
) -> Union[T_Type, Callable[[T_Type], T_Type]]:
    """Decorator to whitelist an object (NamedTuple / dataclass / pydantic model) or
    Enum subclass to be serializable. Various arguments can be passed to alter
    serialization behavior for backcompat purposes.

    Args:
      serializer (Type[Serializer]):
          A custom serializer class to use. Object serializers should inherit from the appropriate
          `ObjectSerializer` subclass and use any of hooks `before_unpack`,
          `after_pack`, and `handle_unpack_error`. For Enums, should inherit from `EnumSerializer`
          and use any of hooks `pack` or `unpack`.
      storage_name (Optional[str]):
          A string that will replace the class name when serializing and deserializing. For example,
          if `StorageFoo` is set as the `storage_name` for `Foo`, then `Foo` will be serialized with
          `"__class__": "StorageFoo"`, and dicts encountered with `"__class__": "StorageFoo"` during
          deserialization will be handled by the `Foo` serializer.
      old_storage_names (Optional[AbstractSet[str]]):
          A set of strings that act as aliases for the target class when deserializing. For example,
          if `OldFoo` is passed as an old storage name for `Foo`, then dicts encountered with
          `"__class__": "OldFoo"` during deserialization will be handled by the `Foo` serializer.
      storage_field_names (Optional[Mapping[str, str]]):
          A mapping of field names to the names used during serializing and deserializing. For
          example, if `{"bar": "baz"}` is passed as `storage_field_names` for `Foo`, then
          serializing `Foo(bar=1)` will give `{"__class__": "Foo", "baz": 1}`, and deserializing
          this dict will give `Foo(bar=1)`. Does not apply to enums.
      old_fields (Optional[Mapping[str, JsonSerializableValue]]):
         A mapping of old field names to default values that will be assigned on serialization. This
         is useful for providing backwards compatibility for fields that have been deleted. For
         example, if `{"bar": None}` is passed as `old_fields` for Foo (which has no defined `bar`
         field), then serializing `Foo(...)` will give `{"__class__": "Foo", "bar": None, ...}`.
         Does not apply to enums.
      skip_when_empty_fields (Optional[AbstractSet[str]]):
          A set of fields that should be skipped when serializing if they are an empty collection
          (list, dict, tuple, or set) or None. This allows a stable snapshot ID to be maintained when a new
          field is added. For example, if `{"bar"}` is passed as `skip_when_empty_fields` for `Foo`,
          then serializing `Foo(bar=[])` will give `{"__class__": "Foo"}` (no `bar` in the
          serialization). This is because the `[]` is an empty list and so is dropped. Does not
          apply to enums.
      skip_when_none_fields (Optional[AbstractSet[str]]):
          A set of fields that should be skipped when serializing if they are None. Unlike `skip_when_empty_fields`,
          this will not skip fields that are empty collections, allowing preservation of a
          distinction between None and empty collections across serialization boundaries. This
          allows a stable snapshot ID to be maintained when a new field is added. For example, if
          `{"bar"}` is passed as `skip_when_none_fields` for `Foo`, then serializing `Foo(bar=None)`
          will give `{"__class__": "Foo"}` (no `bar` in the serialization). This is because the `None`
          is an empty list and so is dropped. Does not apply to enums.
      field_serializers (Optional[Mapping[str, FieldSerializer]]):
          A mapping of field names to `FieldSerializer` classes containing custom serialization
          logic. If a field has an associated `FieldSerializer`, then the `pack` and `unpack`
          methods on the `FieldSerializer` will be used instead of defaults. This is mostly useful
          for data structures that are broadly used throughout the codebase but change format.
          The canonical example is `MetadataFieldSerializer`.  Note that if the field needs to be
          stored under a different name, then it still needs an entry in `storage_field_names` even
          if a `FieldSerializer` is provided. Does not apply to enums.
        kwargs_fields (Optional[AbstractSet[str]]): A set of fields that will be passed to the constructor in kwargs.
    """
    if storage_field_names or old_fields or skip_when_empty_fields or skip_when_none_fields:
        check.invariant(
            serializer is None or issubclass(serializer, ObjectSerializer),
            "storage_field_names, old_fields, skip_when_empty_fields, skip_when_none_fields can only be used with a"
            " ObjectSerializer",
        )
    if __cls is not None:  # decorator invoked directly on class
        check.class_param(__cls, "__cls")
        return _whitelist_for_serdes(
            whitelist_map=_WHITELIST_MAP,
        )(__cls)
    else:  # decorator passed params
        check.opt_class_param(serializer, "serializer", superclass=Serializer)

        return _whitelist_for_serdes(
            whitelist_map=_WHITELIST_MAP,
            serializer=serializer,
            storage_name=storage_name,
            old_storage_names=old_storage_names,
            storage_field_names=storage_field_names,
            old_fields=old_fields,
            skip_when_empty_fields=skip_when_empty_fields,
            skip_when_none_fields=skip_when_none_fields,
            field_serializers=field_serializers,
            kwargs_fields=kwargs_fields,
        )


def _whitelist_for_serdes(
    whitelist_map: WhitelistMap,
    serializer: Optional[type["Serializer"]] = None,
    storage_name: Optional[str] = None,
    old_storage_names: Optional[AbstractSet[str]] = None,
    storage_field_names: Optional[Mapping[str, str]] = None,
    old_fields: Optional[Mapping[str, JsonSerializableValue]] = None,
    skip_when_empty_fields: Optional[AbstractSet[str]] = None,
    skip_when_none_fields: Optional[AbstractSet[str]] = None,
    field_serializers: Optional[Mapping[str, type["FieldSerializer"]]] = None,
    kwargs_fields: Optional[AbstractSet[str]] = None,
) -> Callable[[T_Type], T_Type]:
    def __whitelist_for_serdes(klass: T_Type) -> T_Type:
        if issubclass(klass, Enum) and (
            serializer is None or issubclass(serializer, EnumSerializer)
        ):
            whitelist_map.register_enum(
                klass.__name__,
                klass,
                serializer,
                storage_name=storage_name,
                old_storage_names=old_storage_names,
            )
            return klass
        elif is_named_tuple_subclass(klass) and (
            serializer is None or issubclass(serializer, NamedTupleSerializer)
        ):
            _check_serdes_tuple_class_invariants(
                klass,
                kwargs_fields=kwargs_fields,
            )

            whitelist_map.register_object(
                klass.__name__,
                klass,
                serializer or NamedTupleSerializer,
                storage_name=storage_name,
                old_storage_names=old_storage_names,
                storage_field_names=storage_field_names,
                old_fields=old_fields,
                skip_when_empty_fields=skip_when_empty_fields,
                skip_when_none_fields=skip_when_none_fields,
                field_serializers=field_serializers,
                kwargs_fields=kwargs_fields,
            )
            return klass  # type: ignore  # (NamedTuple quirk)

        elif is_dataclass(klass) and (
            serializer is None or issubclass(serializer, DataclassSerializer)
        ):
            whitelist_map.register_object(
                klass.__name__,
                klass,
                serializer or DataclassSerializer,
                storage_name=storage_name,
                old_storage_names=old_storage_names,
                storage_field_names=storage_field_names,
                old_fields=old_fields,
                skip_when_empty_fields=skip_when_empty_fields,
                skip_when_none_fields=skip_when_none_fields,
                field_serializers=field_serializers,
            )
            return klass  # type: ignore
        else:
            # defer to the last possible moment for import performance
            from pydantic import BaseModel

            if issubclass(klass, BaseModel) and (
                serializer is None or issubclass(serializer, PydanticModelSerializer)
            ):
                whitelist_map.register_object(
                    klass.__name__,
                    klass,
                    serializer or PydanticModelSerializer,
                    storage_name=storage_name,
                    old_storage_names=old_storage_names,
                    storage_field_names=storage_field_names,
                    old_fields=old_fields,
                    skip_when_empty_fields=skip_when_empty_fields,
                    skip_when_none_fields=skip_when_none_fields,
                    field_serializers=field_serializers,
                )
                return klass  # type: ignore
            else:
                raise SerdesUsageError(
                    f"Can not whitelist class {klass} for serializer {serializer}"
                )

    return __whitelist_for_serdes


def is_whitelisted_for_serdes_object(
    val: Any, whitelist_map: WhitelistMap = _WHITELIST_MAP
) -> bool:
    """Check if object has been decorated with `@whitelist_for_serdes`."""
    from pydantic import BaseModel

    if (
        (isinstance(val, tuple) and hasattr(val, "_fields"))  # NamedTuple
        or (is_dataclass(val) and not isinstance(val, type))  # dataclass instance
        or isinstance(val, BaseModel)
    ):
        klass_name = val.__class__.__name__
        return klass_name in whitelist_map.object_serializers

    return False


###################################################################################################
# Serializers
###################################################################################################


class UnpackContext:
    """values are unpacked bottom up."""

    def __init__(self):
        self.observed_unknown_serdes_values: set[UnknownSerdesValue] = set()

    def assert_no_unknown_values(self, obj: UnpackedValue) -> PackableValue:
        if isinstance(obj, UnknownSerdesValue):
            raise DeserializationError(
                f"{obj.message}\nThis error can occur due to version skew, verify processes are"
                " running expected versions."
            )
        elif isinstance(obj, (list, set, frozenset)):
            for inner in obj:
                self.assert_no_unknown_values(inner)
        elif isinstance(obj, dict):
            for v in obj.values():
                self.assert_no_unknown_values(v)

        return cast("PackableValue", obj)

    def observe_unknown_value(self, val: "UnknownSerdesValue") -> "UnknownSerdesValue":
        self.observed_unknown_serdes_values.add(val)
        return val

    def clear_ignored_unknown_values(self, obj: T) -> T:
        if isinstance(obj, UnknownSerdesValue):
            self.observed_unknown_serdes_values.discard(obj)
            self.clear_ignored_unknown_values(obj.value)
        elif isinstance(obj, (list, set, frozenset)):
            for inner in obj:
                self.clear_ignored_unknown_values(inner)
        elif isinstance(obj, dict):
            for v in obj.values():
                self.clear_ignored_unknown_values(v)

        return obj

    def finalize_unpack(self, unpacked: UnpackedValue) -> PackableValue:
        if self.observed_unknown_serdes_values:
            message = ",".join(v.message for v in self.observed_unknown_serdes_values)
            raise DeserializationError(
                f"{message}\nThis error can occur due to version skew, verify processes are"
                " running expected versions."
            )
        return cast("PackableValue", unpacked)


class Serializer(ABC):
    pass


T_Enum = TypeVar("T_Enum", bound=Enum, default=Enum)


class EnumSerializer(Serializer, Generic[T_Enum]):
    def __init__(self, *, klass: type[T_Enum], storage_name: Optional[str] = None):
        self.klass = klass
        self.storage_name = storage_name

    def unpack(self, value: str) -> T_Enum:
        return self.klass[value]

    def pack(
        self,
        value: Enum,
        whitelist_map: WhitelistMap,
        descent_path: str,
    ) -> str:
        return f"{self.get_storage_name()}.{value.name}"

    def get_storage_name(self) -> str:
        return self.storage_name or self.klass.__name__


EMPTY_VALUES_TO_SKIP: tuple[None, list[Any], dict[Any, Any], set[Any]] = (
    None,
    [],
    {},
    set(),
)


class ObjectSerializer(Serializer, Generic[T]):
    # NOTE: See `whitelist_for_serdes` docstring for explanations of parameters.
    def __init__(
        self,
        *,
        klass: type[T],
        storage_name: Optional[str] = None,
        storage_field_names: Optional[Mapping[str, str]] = None,
        old_fields: Optional[Mapping[str, JsonSerializableValue]] = None,
        skip_when_empty_fields: Optional[AbstractSet[str]] = None,
        skip_when_none_fields: Optional[AbstractSet[str]] = None,
        field_serializers: Optional[Mapping[str, "FieldSerializer"]] = None,
        kwargs_fields: Optional[AbstractSet[str]] = None,
    ):
        self.klass = klass
        self.storage_name = storage_name
        self.storage_field_names = storage_field_names or {}
        self.loaded_field_names = {v: k for k, v in self.storage_field_names.items()}
        self.old_fields = old_fields or {}
        self.skip_when_empty_fields = skip_when_empty_fields or set()
        self.skip_when_none_fields = skip_when_none_fields or set()
        self.field_serializers = field_serializers or {}
        self.kwargs_fields = kwargs_fields

    @abstractmethod
    def object_as_mapping(self, value: T) -> Mapping[str, PackableValue]: ...

    def unpack(
        self,
        unpacked_dict: dict[str, UnpackedValue],
        whitelist_map: WhitelistMap,
        context: UnpackContext,
    ) -> T:
        try:
            unpacked_dict = self.before_unpack(context, unpacked_dict)
            unpacked: dict[str, PackableValue] = {}
            for key, value in unpacked_dict.items():
                loaded_name = self.loaded_field_names.get(key, key)
                # Naively implements backwards compatibility by filtering arguments that aren't present in
                # the constructor. If a property is present in the serialized object, but doesn't exist in
                # the version of the class loaded into memory, that property will be completely ignored.
                if loaded_name in self.constructor_param_names:
                    # custom unpack regardless of hook vs recursive descent
                    custom = self.field_serializers.get(loaded_name)
                    if custom:
                        unpacked[loaded_name] = custom.unpack(
                            value,
                            whitelist_map=whitelist_map,
                            context=context,
                        )
                    elif context.observed_unknown_serdes_values:
                        unpacked[loaded_name] = context.assert_no_unknown_values(value)
                    else:
                        unpacked[loaded_name] = value  # type: ignore # 2 hot 4 cast()

                else:
                    context.clear_ignored_unknown_values(value)

            return self.klass(**unpacked)
        except Exception as exc:
            value = self.handle_unpack_error(exc, context, unpacked_dict)
            if isinstance(context, UnpackContext):
                context.assert_no_unknown_values(value)
                context.clear_ignored_unknown_values(unpacked_dict)
            return value

    # Hook: Modify the contents of the unpacked dict before domain object construction during
    # deserialization.
    def before_unpack(
        self,
        context: UnpackContext,
        unpacked_dict: dict[str, UnpackedValue],
    ) -> dict[str, UnpackedValue]:
        return unpacked_dict

    # Hook: Handle an error that occurs when unpacking an object. Can be used to return a default
    # value.
    def handle_unpack_error(
        self,
        exc: Exception,
        context: UnpackContext,
        storage_dict: dict[str, Any],
    ) -> Any:
        raise exc

    def pack_items(
        self,
        value: T,
        whitelist_map: WhitelistMap,
        object_handler: Callable[[SerializableObject, WhitelistMap, str], JsonSerializableValue],
        descent_path: str,
    ) -> Iterator[tuple[str, JsonSerializableValue]]:
        yield "__class__", self.get_storage_name()
        for key, inner_value in self.object_as_mapping(self.before_pack(value)).items():
            if (key in self.skip_when_empty_fields and inner_value in EMPTY_VALUES_TO_SKIP) or (
                key in self.skip_when_none_fields and inner_value is None
            ):
                continue
            storage_key = self.storage_field_names.get(key, key)
            custom = self.field_serializers.get(key)
            if custom:
                yield (
                    storage_key,
                    custom.pack(
                        inner_value,
                        whitelist_map=whitelist_map,
                        descent_path=f"{descent_path}.{key}",
                    ),
                )
            else:
                yield (
                    storage_key,
                    _transform_for_serialization(
                        inner_value,
                        whitelist_map=whitelist_map,
                        object_handler=object_handler,
                        descent_path=f"{descent_path}.{key}",
                    ),
                )
        for key, default in self.old_fields.items():
            yield key, default

    # Hook: Modify the contents of the object before packing
    def before_pack(self, value: T) -> T:
        return value

    @property
    @abstractmethod
    def constructor_param_names(self) -> Sequence[str]: ...

    def get_storage_name(self) -> str:
        return self.storage_name or self.klass.__name__


T_NamedTuple = TypeVar("T_NamedTuple", default=NamedTuple)


# T_NamedTuple previously had a `NamedTuple` bound, but the bound triggers type errors when a `@record`
# decorated class is bound to the variable. @record actually generates a NamedTuple variant of the
# class that was passed in, but we haven't found out how to communicate that to the type system.
# Instead the type signature of the `@record` decorator passes the input class through unmodified.
# Therefore, we forgo `bound` here so that `NamedTupleSerializer` can be used with `@record`
# classes.
class NamedTupleSerializer(ObjectSerializer[T_NamedTuple]):
    def object_as_mapping(self, value: T_NamedTuple) -> Mapping[str, Any]:
        if is_record(value):
            return as_dict_for_new(value)

        # Value is always a NamedTuple, we just can't express that in the type of T_NamedTuple.
        return value._asdict()  # type: ignore

    @cached_property
    def constructor_param_names(self) -> Sequence[str]:  # pyright: ignore[reportIncompatibleMethodOverride]
        if has_generated_new(self.klass):
            return list(get_record_annotations(self.klass).keys())

        names = []
        for name, parameter in signature(self.klass.__new__).parameters.items():
            if parameter.kind is Parameter.VAR_POSITIONAL:
                check.failed("Can not use positional args capture on serdes object.")
            elif parameter.kind is Parameter.VAR_KEYWORD:
                names.extend(
                    check.not_none(
                        self.kwargs_fields,
                        "Must specify kwargs_fields when using kwarg capture in __new__.",
                    )
                )
            else:
                names.append(name)

        return names


# Alias for clarity-- see note on `T_NamedTuple` for the relationship between `NamedTuple` and
# `@record`-decorated classes.
RecordSerializer = NamedTupleSerializer

T_Dataclass = TypeVar("T_Dataclass", bound="DataclassInstance", default="DataclassInstance")


class DataclassSerializer(ObjectSerializer[T_Dataclass]):
    def object_as_mapping(self, value: T_Dataclass) -> Mapping[str, Any]:
        return value.__dict__

    @cached_property
    def constructor_param_names(self) -> Sequence[str]:  # pyright: ignore[reportIncompatibleMethodOverride]
        return list(f.name for f in dataclasses.fields(self.klass))


T_PydanticModel = TypeVar("T_PydanticModel", bound="BaseModel", default="BaseModel")


class PydanticModelSerializer(ObjectSerializer[T_PydanticModel]):
    def object_as_mapping(self, value: T_PydanticModel) -> Mapping[str, Any]:
        value_dict = value.__dict__

        result = {}
        for key, field in self._model_fields.items():
            if field.alias is None and (
                field.serialization_alias is not None or field.validation_alias is not None
            ):
                raise SerializationError(
                    "Can't serialize pydantic models with serialization or validation aliases. Use "
                    "the storage_field_names argument to whitelist_for_serdes instead."
                )
            result_key = field.alias if field.alias else key
            result[result_key] = value_dict[key]

        return result

    @cached_property
    def constructor_param_names(self) -> Sequence[str]:  # pyright: ignore[reportIncompatibleMethodOverride]
        return [field.alias or key for key, field in self._model_fields.items()]

    @cached_property
    def _model_fields(self) -> Mapping[str, Any]:
        # defer for import performance
        from dagster_shared.dagster_model.pydantic_compat_layer import model_fields

        return model_fields(self.klass)


class FieldSerializer(Serializer):
    _instance = None

    # Because `FieldSerializer` is not currently parameterizable (all variation is contained in the
    # logic of pack/unpack), we store references to a singleton instance in the whitelist map to
    # save memory.
    @classmethod
    def get_instance(cls) -> Self:
        if cls._instance is None:
            cls._instance = cls()
        return cls._instance

    @abstractmethod
    def unpack(
        self,
        __unpacked_value: UnpackedValue,
        whitelist_map: WhitelistMap,
        context: UnpackContext,
    ) -> PackableValue: ...

    @abstractmethod
    def pack(
        self,
        __unpacked_value: Any,
        whitelist_map: WhitelistMap,
        descent_path: str,
    ) -> JsonSerializableValue: ...


class SetToSequenceFieldSerializer(FieldSerializer):
    def unpack(  # pyright: ignore[reportIncompatibleMethodOverride]
        self, sequence_value: Optional[Sequence[Any]], **_kwargs
    ) -> Optional[AbstractSet[Any]]:
        return set(sequence_value) if sequence_value is not None else None

    def pack(
        self,
        set_value: Optional[AbstractSet[Any]],
        whitelist_map: WhitelistMap,
        descent_path: str,
    ) -> Optional[Sequence[Any]]:
        return (
            sorted([pack_value(x, whitelist_map, descent_path) for x in set_value], key=str)
            if set_value is not None
            else None
        )


###################################################################################################
# Serialize / Pack
###################################################################################################


def serialize_value(
    val: PackableValue,
    whitelist_map: WhitelistMap = _WHITELIST_MAP,
    **json_kwargs: Any,
) -> str:
    """Serialize an object to a JSON string.

    Objects are first converted to a JSON-serializable form with `pack_value`.
    """
    serializable_value = _transform_for_serialization(
        val,
        whitelist_map=whitelist_map,
        object_handler=_wrap_object,
        descent_path=_root(val),
    )
    return seven.json.dumps(serializable_value, **json_kwargs)


@overload
def pack_value(
    val: T_Scalar,
    whitelist_map: WhitelistMap = ...,
    descent_path: Optional[str] = ...,
) -> T_Scalar: ...


@overload  # for all the types that serialize to JSON object
def pack_value(
    val: Union[
        Mapping[str, PackableValue],
        set[PackableValue],
        frozenset[PackableValue],
        NamedTuple,
        "DataclassInstance",
        "BaseModel",
        Enum,
    ],
    whitelist_map: WhitelistMap = ...,
    descent_path: Optional[str] = ...,
) -> Mapping[str, JsonSerializableValue]: ...


@overload
def pack_value(
    val: Sequence[PackableValue],
    whitelist_map: WhitelistMap = ...,
    descent_path: Optional[str] = ...,
) -> Sequence[JsonSerializableValue]: ...


@overload
def pack_value(
    val: PackableValue,
    whitelist_map: WhitelistMap = ...,
    descent_path: Optional[str] = ...,
) -> JsonSerializableValue: ...


def pack_value(
    val: PackableValue,
    whitelist_map: WhitelistMap = _WHITELIST_MAP,
    descent_path: Optional[str] = None,
) -> JsonSerializableValue:
    """Convert an object into a json serializable complex of dicts, lists, and scalars.

    The following types are transformed in to dicts with special marker keys:
        * whitelisted objects (NamedTuple, dataclass, or pydantic models)
        * whitelisted enums
        * set
        * frozenset
    """
    descent_path = _root(val) if descent_path is None else descent_path
    return _transform_for_serialization(
        val,
        whitelist_map=whitelist_map,
        descent_path=descent_path,
        object_handler=_pack_object,
    )


def _transform_for_serialization(
    val: PackableValue,
    whitelist_map: WhitelistMap,
    object_handler: Callable[[SerializableObject, WhitelistMap, str], JsonSerializableValue],
    descent_path: str,
) -> JsonSerializableValue:
    # this is a hot code path so we handle the common base cases without isinstance
    tval = type(val)
    if tval in (int, float, str, bool) or val is None:
        return val  # type: ignore # 2 hot 4 cast()
    if tval is list:
        return [
            _transform_for_serialization(
                item,
                whitelist_map,
                object_handler,
                f"{descent_path}[{idx}]",
            )
            for idx, item in enumerate(cast("list", val))
        ]
    if tval is dict:
        return {
            key: _transform_for_serialization(
                value,
                whitelist_map,
                object_handler,
                f"{descent_path}.{key}",
            )
            for key, value in cast("dict", val).items()
        }
    if tval is SerializableNonScalarKeyMapping:
        return {
            "__mapping_items__": [
                [
                    _transform_for_serialization(
                        k,
                        whitelist_map,
                        object_handler,
                        f"{descent_path}.{k}",
                    ),
                    _transform_for_serialization(
                        v,
                        whitelist_map,
                        object_handler,
                        f"{descent_path}.{k}",
                    ),
                ]
                for k, v in cast("dict", val).items()
            ]
        }

    if isinstance(val, Enum):
        klass_name = val.__class__.__name__
        if klass_name not in whitelist_map.enum_serializers:
            raise SerializationError(
                "Can only serialize whitelisted Enums, received"
                f" {klass_name}.\nDescent path: {descent_path}",
            )
        enum_serializer = whitelist_map.enum_serializers[klass_name]
        return {"__enum__": enum_serializer.pack(val, whitelist_map, descent_path)}

    # defer to the last possible moment for import performance
    from pydantic import BaseModel

    if (
        (isinstance(val, tuple) and hasattr(val, "_fields"))
        or isinstance(val, BaseModel)
        or (is_dataclass(val) and not isinstance(val, type))
    ):
        klass_name = val.__class__.__name__
        if klass_name not in whitelist_map.object_serializers:
            raise SerializationError(
                "Can only serialize whitelisted namedtuples, received"
                f" {val}.\nDescent path: {descent_path}",
            )
        return object_handler(
            cast("SerializableObject", val),
            whitelist_map,
            descent_path,
        )
    if isinstance(val, set):
        set_path = descent_path + "{}"
        return {
            "__set__": [
                _transform_for_serialization(
                    item,
                    whitelist_map,
                    object_handler,
                    set_path,
                )
                for item in sorted(list(val), key=str)
            ]
        }
    if isinstance(val, frozenset):
        frz_set_path = descent_path + "{}"
        return {
            "__frozenset__": [
                _transform_for_serialization(
                    item,
                    whitelist_map,
                    object_handler,
                    frz_set_path,
                )
                for item in sorted(list(val), key=str)
            ]
        }

    # custom string subclasses
    if isinstance(val, str):
        return val

    # handle more expensive and uncommon abc instance checks last
    if isinstance(val, collections.abc.Mapping):
        return {
            key: _transform_for_serialization(
                value,
                whitelist_map,
                object_handler,
                f"{descent_path}.{key}",
            )
            for key, value in val.items()
        }
    if isinstance(val, collections.abc.Sequence):
        return [
            _transform_for_serialization(
                item,
                whitelist_map,
                object_handler,
                f"{descent_path}[{idx}]",
            )
            for idx, item in enumerate(val)
        ]

    raise SerializationError(f"Unhandled value type {tval}")


def _pack_object(
    obj: SerializableObject, whitelist_map: WhitelistMap, descent_path: str
) -> Mapping[str, JsonSerializableValue]:
    # the object_handler for _transform_for_serialization to produce dicts for objects

    klass_name = obj.__class__.__name__
    serializer = whitelist_map.object_serializers[klass_name]
    return dict(serializer.pack_items(obj, whitelist_map, _pack_object, descent_path))


class _LazySerializationWrapper(dict):
    """An object used to allow us to drive serialization iteratively
    over the tree of objects via json.dumps, instead of having to create
    an entire tree of primitive objects to pass to json.dumps.
    """

    __slots__ = [
        "_descent_path",
        "_obj",
        "_whitelist_map",
    ]

    def __init__(
        self,
        obj: SerializableObject,
        whitelist_map: WhitelistMap,
        descent_path: str,
    ):
        self._obj = obj
        self._whitelist_map = whitelist_map
        self._descent_path = descent_path

        # populate backing native dict to work around c fast path check
        # https://github.com/python/cpython/blob/0fb18b02c8ad56299d6a2910be0bab8ad601ef24/Modules/_json.c#L1542
        super().__init__({"__serdes": "wrapper"})

    def items(self) -> Iterator[tuple[str, JsonSerializableValue]]:  # pyright: ignore[reportIncompatibleMethodOverride]
        klass_name = self._obj.__class__.__name__
        serializer = self._whitelist_map.object_serializers[klass_name]
        yield from serializer.pack_items(
            self._obj, self._whitelist_map, _wrap_object, self._descent_path
        )


def _wrap_object(
    obj: SerializableObject,
    whitelist_map: WhitelistMap,
    descent_path: str,
) -> "_LazySerializationWrapper":
    # the object_handler for _transform_for_serialization to use in conjunction with json.dumps for iterative serialization
    return _LazySerializationWrapper(obj, whitelist_map, descent_path)


###################################################################################################
# Deserialize / Unpack
###################################################################################################

T_PackableValue = TypeVar("T_PackableValue", bound=PackableValue, default=PackableValue)
U_PackableValue = TypeVar("U_PackableValue", bound=PackableValue, default=PackableValue)


@overload
def deserialize_value(
    val: str,
    as_type: tuple[type[T_PackableValue], type[U_PackableValue]],
    whitelist_map: WhitelistMap = ...,
) -> Union[T_PackableValue, U_PackableValue]: ...


@overload
def deserialize_value(
    val: str,
    as_type: type[T_PackableValue],
    whitelist_map: WhitelistMap = ...,
) -> T_PackableValue: ...


@overload
def deserialize_value(
    val: str,
    as_type: None = ...,
    whitelist_map: WhitelistMap = ...,
) -> PackableValue: ...


def deserialize_value(
    val: str,
    as_type: Optional[
        Union[type[T_PackableValue], tuple[type[T_PackableValue], type[U_PackableValue]]]
    ] = None,
    whitelist_map: WhitelistMap = _WHITELIST_MAP,
) -> Union[PackableValue, T_PackableValue, Union[T_PackableValue, U_PackableValue]]:
    """Deserialize a json encoded string to a Python object.

    Two steps:

    - Parse the input string as JSON with an object_hook for custom types.
    - Optionally, check that the resulting object is of the expected type.
    """
    check.str_param(val, "val")

    return deserialize_values([val], as_type, whitelist_map)[0]


@overload
def deserialize_values(
    vals: Iterable[str],
    as_type: type[T_PackableValue],
    whitelist_map: WhitelistMap = ...,
) -> Sequence[T_PackableValue]: ...


@overload
def deserialize_values(
    vals: Iterable[str],
    as_type: None = ...,
    whitelist_map: WhitelistMap = ...,
) -> Sequence[PackableValue]: ...


@overload
def deserialize_values(
    vals: Iterable[str],
    as_type: Optional[
        Union[type[T_PackableValue], tuple[type[T_PackableValue], type[U_PackableValue]]]
    ],
    whitelist_map: WhitelistMap = ...,
) -> Sequence[Union[PackableValue, T_PackableValue, Union[T_PackableValue, U_PackableValue]]]: ...


def deserialize_values(
    vals: Iterable[str],
    as_type: Optional[
        Union[type[T_PackableValue], tuple[type[T_PackableValue], type[U_PackableValue]]]
    ] = None,
    whitelist_map: WhitelistMap = _WHITELIST_MAP,
) -> Sequence[Union[PackableValue, T_PackableValue, Union[T_PackableValue, U_PackableValue]]]:
    """Deserialize a collection of values without having to repeatedly exit/enter the deserializing context."""
    with (
        disable_dagster_warnings(),
        check.EvalContext.contextual_namespace(whitelist_map.object_type_map),
    ):
        unpacked_values = []
        for val in vals:
            context = UnpackContext()
            unpacked_value = seven.json.loads(
                val,
                object_hook=partial(_unpack_object, whitelist_map=whitelist_map, context=context),
            )
            unpacked_value = context.finalize_unpack(unpacked_value)
            is_match = (
                match_type(unpacked_value, as_type)  # type: ignore
                if as_type is not None
                else True
            )
            if as_type and not is_match:
                raise DeserializationError(
                    f"Deserialized object was not expected type {as_type}, got {type(unpacked_value)}"
                )
            unpacked_values.append(unpacked_value)

    return unpacked_values


class UnknownSerdesValue:
    def __init__(self, message: str, value: Mapping[str, UnpackedValue]):
        self.message = message
        self.value = value


def _unpack_object(val: dict, whitelist_map: WhitelistMap, context: UnpackContext) -> UnpackedValue:
    if "__class__" in val:
        klass_name = val["__class__"]
        if klass_name not in whitelist_map.object_deserializers:
            return context.observe_unknown_value(
                UnknownSerdesValue(
                    f'Attempted to deserialize class "{klass_name}" which is not in the whitelist.',
                    val,
                )
            )

        val.pop("__class__")
        deserializer = whitelist_map.object_deserializers[klass_name]
        return deserializer.unpack(val, whitelist_map, context)

    if "__enum__" in val:
        enum = cast("str", val["__enum__"])
        name, member = enum.split(".")
        if name not in whitelist_map.enum_serializers:
            return context.observe_unknown_value(
                UnknownSerdesValue(
                    f"Attempted to deserialize enum {name} which was not in the whitelist.",
                    val,
                )
            )

        enum_serializer = whitelist_map.enum_serializers[name]
        return enum_serializer.unpack(member)

    if "__set__" in val:
        items = cast("list[JsonSerializableValue]", val["__set__"])
        return set(items)

    if "__frozenset__" in val:
        items = cast("list[JsonSerializableValue]", val["__frozenset__"])
        return frozenset(items)

    if "__mapping_items__" in val:
        return {
            cast("Any", _unpack_value(k, whitelist_map, context)): _unpack_value(
                v, whitelist_map, context
            )
            for k, v in val["__mapping_items__"]
        }

    return val


@overload
def unpack_value(
    val: JsonSerializableValue,
    as_type: tuple[type[T_PackableValue], type[U_PackableValue]],
    whitelist_map: WhitelistMap = ...,
    context: Optional[UnpackContext] = ...,
) -> Union[T_PackableValue, U_PackableValue]: ...


@overload
def unpack_value(
    val: JsonSerializableValue,
    as_type: type[T_PackableValue],
    whitelist_map: WhitelistMap = ...,
    context: Optional[UnpackContext] = ...,
) -> T_PackableValue: ...


@overload
def unpack_value(
    val: JsonSerializableValue,
    as_type: None = ...,
    whitelist_map: WhitelistMap = ...,
    context: Optional[UnpackContext] = ...,
) -> PackableValue: ...


def unpack_value(
    val: JsonSerializableValue,
    as_type: Optional[
        Union[type[T_PackableValue], tuple[type[T_PackableValue], type[U_PackableValue]]]
    ] = None,
    whitelist_map: WhitelistMap = _WHITELIST_MAP,
    context: Optional[UnpackContext] = None,
) -> Union[PackableValue, T_PackableValue, Union[T_PackableValue, U_PackableValue]]:
    """Convert a JSON-serializable complex of dicts, lists, and scalars into domain objects.

    Dicts with special keys are processed specially:
    - {"__set__": [...]}: becomes a set()
    - {"__frozenset__": [...]}: becomes a frozenset()
    - {"__enum__": "<class>.<name>"}: becomes an Enum class[name], where `class` is an Enum descendant
    - {"__class__": "<class>", ...}: becomes an instance of the class, where `class` is a
        NamedTuple, dataclass or pydantic model
    """
    context = UnpackContext() if context is None else context
    unpacked_value = _unpack_value(
        val,
        whitelist_map,
        context,
    )
    unpacked_value = context.finalize_unpack(unpacked_value)
    if as_type and not isinstance(unpacked_value, as_type):
        raise DeserializationError(
            f"Unpacked object was not expected type {as_type}, got {type(val)}"
        )
    return unpacked_value


def _unpack_value(
    val: JsonSerializableValue,
    whitelist_map: WhitelistMap,
    context: UnpackContext,
) -> UnpackedValue:
    if isinstance(val, list):
        return [_unpack_value(item, whitelist_map, context) for item in val]

    if isinstance(val, dict):
        unpacked_vals = {k: _unpack_value(v, whitelist_map, context) for k, v in val.items()}
        return _unpack_object(unpacked_vals, whitelist_map, context)

    return val


###################################################################################################
# Validation
###################################################################################################


def _check_serdes_tuple_class_invariants(
    klass: type[NamedTuple],
    kwargs_fields: Optional[AbstractSet[str]],
) -> None:
    # can skip validation on @record generated new
    if has_generated_new(klass):
        return

    sig_params = signature(klass.__new__).parameters
    dunder_new_params = list(sig_params.values())

    cls_param = dunder_new_params[0]

    def _with_header(msg: str) -> str:
        return f"For {klass.__name__}: {msg}"

    if cls_param.name not in {"cls", "_cls"}:
        raise SerdesUsageError(
            _with_header(f'First parameter must be _cls or cls. Got "{cls_param.name}".')
        )

    value_params = dunder_new_params[1:]

    for index, field in enumerate(klass._fields):
        if index >= len(value_params):
            if value_params[-1].kind is Parameter.VAR_KEYWORD:
                if kwargs_fields is None:
                    raise SerdesUsageError(
                        _with_header(
                            "kwargs capture used in __new__ but kwargs_fields was not specified. "
                            "Declare the set of fields processed via kwargs as kwargs_fields "
                            f"Expecting: {list(klass._fields[index:])!r}"
                        )
                    )
                elif field not in kwargs_fields:
                    raise SerdesUsageError(
                        _with_header(
                            f'Expected field "{field}" in kwargs_fields but it was not specified.'
                        )
                    )

            else:
                error_msg = (
                    "Missing parameters to __new__. You have declared fields "
                    "that are not present as parameters to the "
                    "to the __new__ method. In order for "
                    "both serdes serialization and pickling to work, "
                    "these must match or kwargs and kwargs_fields must be used. "
                    f"Missing: {list(klass._fields[index:])!r}"
                )

                raise SerdesUsageError(_with_header(error_msg))

        if not is_record(klass):
            # non @record NamedTuples get pickled via ordinal args, so ensure ordering
            value_param = value_params[index]
            if value_param.name != field:
                error_msg = (
                    "Params to __new__ must match the order of field declaration for "
                    "NamedTuples that are not @record based. "
                    f'Declared field number {index + 1} in the namedtuple is "{field}". '
                    f'Parameter {index + 1} in __new__ method is "{value_param.name}".'
                )
                raise SerdesUsageError(_with_header(error_msg))

    if len(value_params) > len(klass._fields):
        # Ensure that remaining parameters have default values
        for extra_param_index in range(len(klass._fields), len(value_params) - 1):
            if value_params[extra_param_index].default == Parameter.empty:
                error_msg = (
                    f'Parameter "{value_params[extra_param_index].name}" is a parameter to the __new__ '
                    "method but is not a field in this namedtuple. The only "
                    "reason why this should exist is that "
                    "it is a field that used to exist (we refer to this as the graveyard) "
                    "but no longer does. However it might exist in historical storage. This "
                    "parameter existing ensures that serdes continues to work. However these "
                    "must come at the end and have a default value for pickling to work."
                )
                raise SerdesUsageError(_with_header(error_msg))


###################################################################################################
# Utilities
###################################################################################################


def _root(val: Any) -> str:
    return f"<root:{val.__class__.__name__}>"


def get_storage_name(klass: type, *, whitelist_map: WhitelistMap = _WHITELIST_MAP):
    if klass.__name__ not in whitelist_map.object_serializers:
        check.failed(f"{klass.__name__} is not a known serializable object type.")
    ser = whitelist_map.object_serializers[klass.__name__]
    return ser.get_storage_name()


def get_storage_fields(klass: type, whitelist_map: WhitelistMap = _WHITELIST_MAP):
    if klass.__name__ not in whitelist_map.object_serializers:
        check.failed(f"{klass.__name__} is not a known serializable object type.")
    ser = whitelist_map.object_serializers[klass.__name__]
    # get the current fields
    current_fields = ser.constructor_param_names
    # remap defined storage field names
    stored_fields = [ser.storage_field_names.get(f, f) for f in current_fields]
    # insert old fields
    stored_fields.extend(ser.old_fields.keys())
    # we sort_keys=True in json dump
    return sorted(stored_fields)


_OBJECT_START = '{"__class__":'


def get_prefix_for_a_serialized(klass: type, *, whitelist_map=_WHITELIST_MAP):
    """Returns the expected start of the string for a serdes serialized object
    of the passed type.
    """
    return f'{_OBJECT_START} "{get_storage_name(klass, whitelist_map=whitelist_map)}"'



from dagster_shared.serdes.objects.package_entry import (
    ComponentFeatureData as ComponentFeatureData,
    PluginObjectKey as PluginObjectKey,
    PluginObjectSnap as PluginObjectSnap,
    ScaffoldTargetTypeData as ScaffoldTargetTypeData,
)



import json
import textwrap
from abc import ABC, abstractmethod
from collections.abc import Sequence
from itertools import groupby
from typing import Any, Literal, Optional, TypedDict, overload

from typing_extensions import TypeAlias

from dagster_shared.record import record
from dagster_shared.serdes.serdes import whitelist_for_serdes


def _generate_invalid_component_typename_error_message(typename: str) -> str:
    return textwrap.dedent(f"""
        Invalid component type name: `{typename}`.
        Type names must be a "."-separated string of valid Python identifiers with at least two segments.
    """)


@whitelist_for_serdes
@record(kw_only=False)
class PluginObjectKey:
    namespace: str
    name: str

    @property
    def package(self) -> str:
        return self.namespace.split(".")[0]

    def to_typename(self) -> str:
        return f"{self.namespace}.{self.name}"

    @staticmethod
    def from_typename(typename: str) -> "PluginObjectKey":
        parts = typename.split(".")
        for part in parts:
            if not part.isidentifier():
                raise ValueError(_generate_invalid_component_typename_error_message(typename))
        if len(parts) < 2:
            raise ValueError(_generate_invalid_component_typename_error_message(typename))
        namespace, _, name = typename.rpartition(".")
        return PluginObjectKey(name=name, namespace=namespace)


###########
# TYPE DATA
###########
PluginObjectFeature: TypeAlias = Literal["component", "scaffold-target"]


class PluginObjectFeatureData(ABC):
    @property
    @abstractmethod
    def feature(self) -> PluginObjectFeature:
        pass


@whitelist_for_serdes
@record
class ComponentFeatureData(PluginObjectFeatureData):
    schema: Optional[dict[str, Any]]

    @property
    def feature(self) -> PluginObjectFeature:
        return "component"


@whitelist_for_serdes
@record
class ScaffoldTargetTypeData(PluginObjectFeatureData):
    schema: Optional[dict[str, Any]]

    @property
    def feature(self) -> PluginObjectFeature:
        return "scaffold-target"


###############
# PLUGIN MANIFEST
###############


@whitelist_for_serdes
@record
class PluginObjectSnap:
    key: PluginObjectKey
    summary: Optional[str]
    description: Optional[str]
    owners: Optional[Sequence[str]]
    tags: Optional[Sequence[str]]
    feature_data: Sequence[PluginObjectFeatureData]

    @property
    def features(self) -> Sequence[PluginObjectFeature]:
        return [type_data.feature for type_data in self.feature_data]

    @overload
    def get_feature_data(self, feature: Literal["component"]) -> Optional[ComponentFeatureData]: ...

    @overload
    def get_feature_data(
        self, feature: Literal["scaffold-target"]
    ) -> Optional[ScaffoldTargetTypeData]: ...

    def get_feature_data(self, feature: PluginObjectFeature) -> Optional[PluginObjectFeatureData]:
        for feature_data in self.feature_data:
            if feature_data.feature == feature:
                return feature_data
        return None

    @property
    def scaffolder_schema(self) -> Optional[dict[str, Any]]:
        scaffolder_data = self.get_feature_data("scaffold-target")
        return scaffolder_data.schema if scaffolder_data else None

    @property
    def component_schema(self) -> Optional[dict[str, Any]]:
        component_data = self.get_feature_data("component")
        return component_data.schema if component_data else None


@whitelist_for_serdes
@record
class PluginManifest:
    """A manifest of all components in a package.

    This is used to generate the component registry and to validate that the package entry point
    is valid.
    """

    modules: Sequence[str]  # List of modules scanned
    objects: Sequence[PluginObjectSnap]

    def merge(self, other: "PluginManifest") -> "PluginManifest":
        """Merge another manifest with this one and return a new instance."""
        shared_modules = set(self.modules).intersection(other.modules)
        if shared_modules:
            raise ValueError(f"Cannot merge manifests with overlapping modules: {shared_modules}.")
        return PluginManifest(
            modules=[*self.modules, *other.modules],
            objects=[*self.objects, *other.objects],
        )


###################################
# COMPONENT REPRESENTATION FOR DOCS
###################################


class ComponentTypeJson(TypedDict):
    """Component type JSON, used to back dg docs webapp."""

    name: str
    owners: Optional[Sequence[str]]
    tags: Optional[Sequence[str]]
    example: str
    schema: str
    description: Optional[str]


class ComponentTypeNamespaceJson(TypedDict):
    """Component type namespace JSON, used to back dg docs webapp."""

    name: str
    componentTypes: list[ComponentTypeJson]


def json_for_all_components(
    components: Sequence[PluginObjectSnap],
) -> list[ComponentTypeNamespaceJson]:
    """Returns a list of JSON representations of all component types in the registry."""
    component_json = []
    for entry in components:
        key = entry.key
        component_type_data = entry.get_feature_data("component")
        if component_type_data:
            component_json.append(
                (
                    key.namespace.split(".")[0],
                    json_for_component_type(key, entry, component_type_data),
                )
            )
    return [
        ComponentTypeNamespaceJson(
            name=namespace,
            componentTypes=[namespace_and_component[1] for namespace_and_component in components],
        )
        for namespace, components in groupby(component_json, key=lambda x: x[0])
    ]


def json_for_component_type(
    key: PluginObjectKey,
    entry: PluginObjectSnap,
    component_type_data: ComponentFeatureData,
) -> ComponentTypeJson:
    from dagster_shared.yaml_utils.sample_yaml import generate_sample_yaml

    typename = key.to_typename()
    sample_yaml = generate_sample_yaml(typename, component_type_data.schema or {})
    return ComponentTypeJson(
        name=typename,
        owners=entry.owners,
        tags=entry.tags,
        example=sample_yaml,
        schema=json.dumps(component_type_data.schema, sort_keys=True),
        description=entry.description,
    )



from typing import Optional, Union

from typing_extensions import TypeAlias

from dagster_shared.record import record
from dagster_shared.serdes import whitelist_for_serdes


@whitelist_for_serdes
@record
class DgAssetMetadata:
    key: str
    deps: list[str]
    kinds: list[str]
    group: Optional[str]
    description: Optional[str]
    automation_condition: Optional[str]


@whitelist_for_serdes
@record
class DgSensorMetadata:
    name: str


@whitelist_for_serdes
@record
class DgScheduleMetadata:
    name: str
    cron_schedule: str


@whitelist_for_serdes
@record
class DgJobMetadata:
    name: str


@whitelist_for_serdes
@record
class DgAssetCheckMetadata:
    key: str
    asset_key: str
    name: str
    additional_deps: list[str]
    description: Optional[str]


DgDefinitionMetadata: TypeAlias = Union[
    DgAssetMetadata,
    DgSensorMetadata,
    DgScheduleMetadata,
    DgJobMetadata,
    DgAssetCheckMetadata,
]



import os
from collections.abc import Mapping
from dataclasses import dataclass
from pathlib import Path
from typing import Any, NamedTuple, Optional

from dagster_shared.merger import deep_merge_dicts
from dagster_shared.utils.config import get_dg_config_path, load_config, write_config

DEFAULT_CLOUD_CLI_FOLDER = os.path.join(os.path.expanduser("~"), ".dagster_cloud_cli")
DEFAULT_CLOUD_CLI_CONFIG = os.path.join(DEFAULT_CLOUD_CLI_FOLDER, "config")


class DagsterPlusConfigInfo(NamedTuple):
    path: Path
    raw_config: Mapping[str, Any]
    plus_config: Mapping[str, Any]
    is_dg_config: bool


DAGSTER_CLOUD_BASE_URL = "https://dagster.cloud"


def _get_dagster_plus_config_path_and_raw_config() -> Optional[DagsterPlusConfigInfo]:
    cloud_config_path = get_dagster_cloud_cli_config_path()
    dg_config_path = get_dg_config_path()

    dg_config = load_config(dg_config_path) if dg_config_path.exists() else None
    dg_plus_config = dg_config.get("cli", {}).get("plus", {}) if dg_config else None
    cloud_config = load_config(cloud_config_path) if cloud_config_path.exists() else None

    if dg_plus_config and cloud_config:
        raise Exception(
            f"Found Dagster Plus config in both {dg_config_path} and {cloud_config_path}. Please consolidate your config files."
        )

    if cloud_config is not None:
        return DagsterPlusConfigInfo(
            cloud_config_path, cloud_config, cloud_config, is_dg_config=False
        )
    elif dg_config is not None and dg_plus_config is not None:
        return DagsterPlusConfigInfo(dg_config_path, dg_config, dg_plus_config, is_dg_config=True)

    return None


@dataclass()
class DagsterPlusCliConfig:
    url: Optional[str] = None
    organization: Optional[str] = None
    default_deployment: Optional[str] = None
    user_token: Optional[str] = None
    agent_timeout: Optional[int] = None

    def has_any_configuration(self) -> bool:
        return any(self.__dict__.values())

    @staticmethod
    def exists() -> bool:
        return _get_dagster_plus_config_path_and_raw_config() is not None

    @classmethod
    def get(cls) -> "DagsterPlusCliConfig":
        result = _get_dagster_plus_config_path_and_raw_config()
        if result is None:
            raise Exception("No Dagster Plus config found")
        _, _, raw_plus_config, _ = result
        return cls(**raw_plus_config)

    def write(self):
        existing_config = _get_dagster_plus_config_path_and_raw_config()
        if existing_config is None:
            config_path = get_dg_config_path()
            raw_config = {}
            is_dg_config = True
        else:
            config_path, raw_config, _, is_dg_config = existing_config

        config_to_apply = {k: v for k, v in self.__dict__.items() if v is not None}
        if is_dg_config:
            config_to_apply = {"cli": {"plus": config_to_apply}}

        config_dict = deep_merge_dicts(raw_config, config_to_apply)
        write_config(config_path, config_dict)

    @property
    def organization_url(self) -> str:
        if not self.organization:
            raise Exception("Organization not set")
        if self.url is None:
            return f"{DAGSTER_CLOUD_BASE_URL}/{self.organization}"
        return f"{self.url}/{self.organization}"


def get_dagster_cloud_cli_config_path() -> Path:
    return Path(os.getenv("DAGSTER_CLOUD_CLI_CONFIG", DEFAULT_CLOUD_CLI_CONFIG))



import _thread as thread
import os
import signal
import subprocess
import sys
import tempfile
import threading
from collections.abc import Iterator, Sequence
from contextlib import contextmanager
from subprocess import Popen
from typing import Any, Callable, Optional

import dagster_shared.check as check
from dagster_shared.seven import IS_WINDOWS

# Windows subprocess termination utilities. See here for why we send on Windows:
# https://stefan.sofa-rockers.org/2013/08/15/handling-sub-process-hierarchies-python-linux-os-x/


def open_ipc_subprocess(parts: Sequence[str], **kwargs: Any) -> "Popen[Any]":
    """Sets the correct flags to support graceful termination."""
    check.list_param(parts, "parts", str)

    creationflags = 0
    if sys.platform == "win32":
        creationflags = subprocess.CREATE_NEW_PROCESS_GROUP

        # pass_fds is not supported on Windows. Instead we set close_fds to False, which will allow
        # any inheritable file descriptors marked as inheritable to be inherited by the child
        # process.
        if kwargs.get("pass_fds"):
            del kwargs["pass_fds"]
            kwargs["close_fds"] = False

    return subprocess.Popen(
        parts,
        creationflags=creationflags,
        **kwargs,
    )


def interrupt_ipc_subprocess(proc: "Popen[Any]", sig: Optional[int] = None) -> None:
    """Send CTRL_BREAK on Windows, SIGINT on other platforms."""
    proc.send_signal(sig or get_default_interrupt_signal())


def interrupt_then_kill_ipc_subprocess(proc: "Popen[Any]", wait_time: int = 10) -> None:
    interrupt_ipc_subprocess(proc)
    try:
        proc.wait(timeout=wait_time)
    except subprocess.TimeoutExpired:
        proc.kill()


def interrupt_ipc_subprocess_pid(pid: int, sig: Optional[int] = None) -> None:
    """Send CTRL_BREAK_EVENT on Windows, SIGINT on other platforms."""
    os.kill(pid, sig or get_default_interrupt_signal())


def get_default_interrupt_signal() -> int:
    return signal.CTRL_BREAK_EVENT if sys.platform == "win32" else signal.SIGINT


def send_interrupt() -> None:
    if IS_WINDOWS:
        # This will raise a KeyboardInterrupt in python land - meaning this wont be able to
        # interrupt things like sleep()
        thread.interrupt_main()
    else:
        # If on unix send an os level signal to interrupt any situation we may be stuck in
        os.kill(os.getpid(), signal.SIGINT)


# ########################
# ##### SHUTDOWN PIPE
# ########################

_PIPE_SHUTDOWN_INDICATOR = "SHUTDOWN"


def get_ipc_shutdown_pipe() -> tuple[int, int]:
    r_fd, w_fd = os.pipe()

    # On windows, convert fd to a Windows handle so it can be reliably passed across processes.
    if sys.platform == "win32":
        import msvcrt

        os.set_inheritable(r_fd, True)
        r_fd = msvcrt.get_osfhandle(r_fd)
    return r_fd, w_fd


@contextmanager
def monitor_ipc_shutdown_pipe(pipe_fd: int, handler: Callable[[], None]) -> Iterator[None]:
    """Monitor the passed in pipe file descriptor for the shutdown indicator message.
    When received, trigger the handler.

    Args:
        pipe_fd: The file descriptor of the pipe to monitor. Must be readable.
            If on windows, this is assumed to be a Windows handle rather than a regular file
            descriptor.
        handler: The handler to call when the shutdown indicator is received.
    """
    # On windows, we expect to receive a raw Windows handle rather than a regular file descriptor.
    # Convert to a file descriptor before reading.
    if sys.platform == "win32":
        import msvcrt

        pipe_fd = msvcrt.open_osfhandle(pipe_fd, os.O_RDONLY)

    break_event = threading.Event()

    def _watch_pipe_for_shutdown():
        with open(pipe_fd) as pipe:
            while not break_event.is_set():
                line = pipe.readline()
                if not line:  # EOF or pipe closed
                    break_event.set()
                elif _PIPE_SHUTDOWN_INDICATOR in line.strip():
                    break_event.set()
                    handler()

    # Start a background thread that watches the pipe
    monitor_thread = threading.Thread(target=_watch_pipe_for_shutdown, daemon=True)
    monitor_thread.start()

    try:
        yield
    finally:
        # Signal the thread to exit and wait for it to stop
        break_event.set()
        monitor_thread.join()


@contextmanager
def interrupt_on_ipc_shutdown_message(pipe_fd: int) -> Iterator[None]:
    """Monitor the passed in pipe file descriptor for the shutdown indicator message. Interrupt the
    current process when the message is received.

    Args:
        pipe_fd: The file descriptor of the pipe to monitor. Must be readable.
            If on windows, this is assumed to be raw Windows handle rather than a regular file
            descriptor.
    """
    # Important to use `send_interrupt` here rather than unconditionally sending a signal. Sending a
    # signal, even to the process itself, often has strange behavior on windows.
    with monitor_ipc_shutdown_pipe(pipe_fd, handler=lambda: send_interrupt()):
        yield


def send_ipc_shutdown_message(w_fd: int) -> None:
    os.write(w_fd, f"{_PIPE_SHUTDOWN_INDICATOR}\n".encode())
    os.close(w_fd)


# ########################
# ##### TEMPFILE
# ########################


@contextmanager
def ipc_tempfile() -> Iterator[str]:
    """Yield a temporary file suitable for cross-platform IPC."""
    # Use delete=False and immediately close for windows compatibility, otherwise windows
    # file-locking will prevent writing to the file.
    temp_file = tempfile.NamedTemporaryFile(delete=False)
    temp_file.close()
    try:
        yield temp_file.name
    finally:
        os.remove(temp_file.name)



from collections.abc import Sequence
from typing import Any, ForwardRef, Literal, NamedTuple, TypeVar, Union, get_args, get_origin

from typing_extensions import TypeGuard

T = TypeVar("T", bound=Any)


# Use Any for type_ because there isn't a good static type that can handle all the cases listed
# here.
def match_type(obj: object, type_: Union[type[T], tuple[type[T]]]) -> TypeGuard[T]:
    if isinstance(type_, tuple):
        return any(match_type(obj, t) for t in type_)

    origin = get_origin(type_)
    # If typ is not a generic alias, do a normal isinstance check
    if origin is None:
        # If we get here, it's a concrete type like `int`, `str`, or type(None).
        if type_ is Any:
            return True
        elif type_ is NamedTuple:
            return is_named_tuple_instance(obj)
        elif isinstance(type_, ForwardRef):
            raise NotImplementedError(
                f"Got ForwardRef {type_}. ForwardRef is not supported by `match_type`"
            )
        else:
            return isinstance(obj, type_)

    # Handle Union (e.g. Union[int, str])
    if origin is Union:
        subtypes = get_args(type_)  # e.g. (int, str)
        return any(match_type(obj, st) for st in subtypes)

    # Handle Literal (e.g. Literal[3, 5, "hello"])
    if origin is Literal:
        # get_args(typ) will be the allowed literal values
        allowed_values = get_args(type_)  # e.g. (3, 5, "hello")
        return obj in allowed_values

    # Handle list[...] (e.g. list[str])
    if origin is Sequence:
        (item_type,) = get_args(type_)  # e.g. (str,) for list[str]
        if not isinstance(obj, Sequence):
            return False
        return all(match_type(item, item_type) for item in obj)

    # Handle list[...] (e.g. list[str])
    if origin is list:
        (item_type,) = get_args(type_)  # e.g. (str,) for list[str]
        if not isinstance(obj, list):
            return False
        return all(match_type(item, item_type) for item in obj)

    # Handle tuple[...] (e.g. tuple[int, str], tuple[str, ...])
    if origin is tuple:
        arg_types = get_args(type_)
        if not isinstance(obj, tuple):
            return False
        # Distinguish fixed-length vs variable-length (ellipsis) tuples
        if len(arg_types) == 2 and arg_types[1] is Ellipsis:
            # e.g. tuple[str, ...]
            elem_type = arg_types[0]
            return all(match_type(item, elem_type) for item in obj)
        else:
            # e.g. tuple[int, str, float]
            if len(obj) != len(arg_types):
                return False
            return all(match_type(item, t) for item, t in zip(obj, arg_types))

    # Handle dict[...] (e.g. dict[str, int])
    if origin is dict:
        key_type, val_type = get_args(type_)
        if not isinstance(obj, dict):
            return False
        return all(match_type(k, key_type) and match_type(v, val_type) for k, v in obj.items())

    # Extend with other generic types (set, frozenset, etc.) if needed
    raise NotImplementedError(f"No handler for {type_}")


# copied from dagster._utils
def is_named_tuple_instance(obj: object) -> TypeGuard[NamedTuple]:
    return isinstance(obj, tuple) and hasattr(obj, "_fields")


# copied from dagster._utils
def is_named_tuple_subclass(klass: type[object]) -> TypeGuard[type[NamedTuple]]:
    return isinstance(klass, type) and issubclass(klass, tuple) and hasattr(klass, "_fields")



"""Internal py2/3 compatibility library. A little more than six."""

import inspect
import os
import shlex
import sys
import time
from collections.abc import Sequence
from pathlib import Path
from types import ModuleType
from typing import Any, Callable, List, Type, Union  # noqa: F401, UP035

from typing_extensions import TypeGuard

import dagster_shared.seven.json as json  # noqa: F401
from dagster_shared.seven.json import (
    JSONDecodeError as JSONDecodeError,
    dump as dump,
    dumps as dumps,
)
from dagster_shared.seven.temp_dir import get_system_temp_directory as get_system_temp_directory

IS_WINDOWS = os.name == "nt"
IS_PYTHON_3_12 = (sys.version_info[0], sys.version_info[1]) == (3, 12)

# TODO implement a generic import by name -- see https://stackoverflow.com/questions/301134/how-to-import-a-module-given-its-name


# https://stackoverflow.com/a/67692/324449
def import_module_from_path(module_name: str, path_to_file: Union[str, Path]) -> ModuleType:
    import importlib.util

    spec = importlib.util.spec_from_file_location(module_name, path_to_file)
    if spec is None:
        raise Exception(
            f"Can not import module {module_name} from path {path_to_file}, unable to load spec."
        )

    if sys.modules.get(spec.name) and spec.origin:
        f = sys.modules[spec.name].__file__
        # __file__ can be relative depending on current working directory
        if f and os.path.abspath(f) == os.path.abspath(spec.origin):
            return sys.modules[spec.name]

    module = importlib.util.module_from_spec(spec)
    sys.modules[spec.name] = module
    assert spec.loader
    spec.loader.exec_module(module)
    return module


def import_uncached_module_from_path(module_name: str, path_to_file: str) -> ModuleType:
    import importlib.util

    spec = importlib.util.spec_from_file_location(module_name, path_to_file)
    if spec is None:
        raise Exception(
            f"Can not import module {module_name} from path {path_to_file}, unable to load spec."
        )

    module = importlib.util.module_from_spec(spec)
    assert spec.loader
    spec.loader.exec_module(module)
    return module


time_fn = time.perf_counter


def get_arg_names(callable_: Callable[..., Any]) -> Sequence[str]:
    return [
        parameter.name
        for parameter in inspect.signature(callable_).parameters.values()
        if parameter.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD
    ]


# https://stackoverflow.com/a/58437485/324449
def is_module_available(module_name: str) -> bool:
    # python 3.4 and above
    import importlib.util

    loader = importlib.util.find_spec(module_name)

    return loader is not None


def is_lambda(target: object) -> TypeGuard[Callable[..., Any]]:
    return callable(target) and getattr(target, "__name__", None) == "<lambda>"


def is_function_or_decorator_instance_of(
    target: object, kls: type[Any]
) -> TypeGuard[Callable[..., Any]]:
    return inspect.isfunction(target) or (isinstance(target, kls) and hasattr(target, "__name__"))


def qualname_differs(target: object) -> bool:
    return hasattr(target, "__qualname__") and getattr(target, "__qualname__") != getattr(
        target, "__name__"
    )


def xplat_shlex_split(s: str) -> list[str]:
    if IS_WINDOWS:
        return shlex.split(s, posix=False)

    return shlex.split(s)


def get_import_error_message(import_error: ImportError) -> str:
    return import_error.msg


def is_subclass(child_type: type[Any], parent_type: type[Any]):
    """Due to some pathological interactions betwen bugs in the Python typing library
    (https://github.com/python/cpython/issues/88459 and
    https://github.com/python/cpython/issues/89010), some types (list[str] in Python 3.9, for
    example) pass inspect.isclass check above but then raise an exception if issubclass is called
    with the same class. This function provides a workaround for that issue.
    """
    if not inspect.isclass(child_type):
        return False

    try:
        return issubclass(child_type, parent_type)
    except TypeError:
        return False



import os
import sys
from pathlib import Path
from typing import Any


def is_windows() -> bool:
    return sys.platform == "win32"


def get_default_dg_user_config_path() -> Path:
    if is_windows():
        return Path.home() / "AppData" / "dg" / "dg.toml"
    else:
        config_home = Path(os.getenv("XDG_CONFIG_HOME", os.path.expanduser("~/.config")))
        return config_home / "dg.toml"


def load_config(config_path: Path) -> dict[str, Any]:
    import tomlkit
    import yaml

    if config_path.suffix == ".toml":
        return tomlkit.parse(config_path.read_text()).unwrap()
    else:
        return yaml.safe_load(config_path.read_text()) or {}


def write_config(config_path: Path, config: dict[str, Any]):
    import tomlkit
    import yaml

    if config_path.suffix == ".toml":
        with open(config_path, "w") as f:
            tomlkit.dump(config, f)
    else:
        with open(config_path, "w") as f:
            yaml.safe_dump(config, f)


def get_dg_config_path() -> Path:
    return Path(os.getenv("DG_CLI_CONFIG", get_default_dg_user_config_path()))


def does_dg_config_file_exist() -> bool:
    return get_dg_config_path().exists()



from collections.abc import Mapping, Sequence
from pathlib import Path
from typing import Any, NamedTuple, Optional, Union, cast

import dagster_shared.check as check

KeyPathSegment = Union[str, int]
KeyPath = Sequence[KeyPathSegment]


class LineCol(NamedTuple):
    line: int
    col: int


class SourcePosition(NamedTuple):
    filename: str
    start: LineCol
    end: LineCol

    def __str__(self):
        return f"{self.filename}:{self.start.line}"


class YamlSourcedError(NamedTuple):
    file_name: str
    start_line_no: int
    location: str
    snippet: str


OFFSET_LINES_BEFORE = 2
OFFSET_LINES_AFTER = 3


class SourcePositionTree(NamedTuple):
    """Represents a tree where every node has a SourcePosition."""

    position: SourcePosition
    children: Mapping[KeyPathSegment, "SourcePositionTree"]

    def __hash__(self) -> int:
        return hash((self.position, tuple(sorted(self.children.items()))))

    def lookup(self, key_path: KeyPath) -> Optional[SourcePosition]:
        """Returns the source position of the descendant at the given path."""
        if len(key_path) == 0:
            return self.position
        head, *tail = key_path
        if head not in self.children:
            return None
        return self.children[head].lookup(tail)

    def lookup_closest_and_path(
        self, key_path: KeyPath, trace: Optional[Sequence[SourcePosition]]
    ) -> tuple[SourcePosition, Sequence[SourcePosition]]:
        """Returns the source position of the descendant at the given path. If the path does not
        exist, returns the source position of the nearest ancestor.
        """
        trace = [*trace, self.position] if trace else [self.position]
        if len(key_path) == 0:
            return self.position, trace
        head, *tail = key_path
        if head not in self.children:
            return self.position, trace
        return self.children[head].lookup_closest_and_path(tail, trace)

    def source_error(
        self,
        yaml_path: KeyPath,
        inline_error_message: str,
        value_error,
    ):
        source_position, source_position_path = self.lookup_closest_and_path(yaml_path, trace=None)

        # Retrieves dotted path representation of the location of the error in the YAML file, e.g.
        # attributes.nested.foo.an_int
        location = ".".join(str(p) for p in yaml_path).split(" at ")[0]

        # Find the first source position that has a different start line than the current source position
        # This is e.g. the parent json key of the current source position
        preceding_source_position = next(
            iter(
                [
                    value
                    for value in reversed(list(source_position_path))
                    if value.start.line < source_position.start.line
                ]
            ),
            source_position,
        )
        if source_position.filename == "<string>":
            code_snippet = ""
        else:
            with open(source_position.filename) as f:
                lines = f.readlines()
            lines_with_line_numbers = list(zip(range(1, len(lines) + 1), lines))

            pre_error_lines = lines_with_line_numbers[
                max(
                    0, preceding_source_position.start.line - OFFSET_LINES_BEFORE
                ) : source_position.start.line
            ]
            source_idx = source_position.start.line - 1
            source_line = (
                lines_with_line_numbers[source_idx][1]
                # parse errors near end of file can spill out of bounds
                if source_idx < len(lines_with_line_numbers)
                else ""
            )
            error_ptr = (
                None,
                _format_indented_error_msg(
                    value_error,
                    source_position,
                    source_line,
                    inline_error_message,
                ),
            )
            post_error_lines = lines_with_line_numbers[
                source_position.start.line : source_position.end.line + OFFSET_LINES_AFTER
            ]

            # Combine the filtered lines with the line numbers, and add empty lines before and after
            lines_with_line_numbers = _prepend_lines_with_line_numbers(
                [(None, ""), *pre_error_lines, error_ptr, *post_error_lines, (None, "")]
            )
            code_snippet = "\n".join(lines_with_line_numbers)

        return YamlSourcedError(
            file_name=source_position.filename,
            start_line_no=source_position.start.line,
            location=location,
            snippet=code_snippet,
        )


def _format_indented_error_msg(
    value_error: bool,
    source_pos: SourcePosition,
    source_line: str,
    msg: str,
) -> str:
    """Format an error message with a caret pointing to the column where the error occurred."""
    inset = source_pos.start.col
    if value_error:
        source_key_end_col = source_pos.end.col
        source_line_end_col = len(source_line)

        approx_middle = source_key_end_col + int((source_line_end_col - source_key_end_col) / 2)
        if approx_middle > source_pos.start.col:  # sanity check we got a reasonable value
            inset = approx_middle

    return " " * (inset - 1) + f"^ {msg}"


def _prepend_lines_with_line_numbers(
    lines_with_numbers: Sequence[tuple[Optional[int], str]],
) -> Sequence[str]:
    """Prepend each line with a line number, right-justified to the maximum line number length.

    Args:
        lines_with_numbers: A sequence of tuples, where the first element is the line number and the
            second element is the line content. Some lines may have a `None` line number, which
            will be rendered as an empty string, used for e.g. inserted error message lines.
    """
    max_line_number_length = max([len(str(n)) for n, _ in lines_with_numbers])
    return [
        f"{(str(n) if n else '').rjust(max_line_number_length)} | {line.rstrip()}"
        for n, line in lines_with_numbers
    ]


class ValueAndSourcePositionTree(NamedTuple):
    """A tree-like object (like a JSON-structured dict) and an accompanying SourcePositionTree.

    Each tree node in the SourcePositionTree is expected to correspond to in value.
    """

    value: Any
    source_position_tree: SourcePositionTree


class SourcePositionAndKeyPath(NamedTuple):
    """Represents a source position and key path within a file.

    Args:
        key_path (KeyPath): The path of keys that lead to the current object, where each element in
            the path is either a string (for dict keys) or an integer (for list indices).
        source_position (Optional[SourcePosition]): The source position of the object in the
            document, if available.
    """

    key_path: KeyPath
    source_position: Optional[SourcePosition]


class HasSourcePositionAndKeyPath:
    _source_position_and_key_path: Optional[SourcePositionAndKeyPath] = None

    @property
    def source_position(self) -> SourcePosition:
        """Returns the underlying source position of the object, including
        the source file and line number.
        """
        assert self._source_position_and_key_path
        assert self._source_position_and_key_path.source_position
        return self._source_position_and_key_path.source_position

    @property
    def source_file(self) -> Path:
        """Path to the source file where the object is defined."""
        assert self._source_position_and_key_path
        return Path(self.source_position.filename)

    @property
    def source_file_name(self) -> str:
        """Name of the source file where the object is defined."""
        return self.source_file.name


def populate_source_position_and_key_paths(
    obj: Any,
    source_position_tree: Optional[SourcePositionTree],
    key_path: KeyPath = [],
) -> None:
    """Populate the SourcePositionAndKeyPath for the given object and its children.
    This function recursively traverses the object and its children, setting the
    SourcePositionAndKeyPath on each object that subclasses HasSourcePositionAndKeyPath.
    If the obj is a collection, its children are the elements in the collection. If obj is an
    object, its children are its attributes.
    The SourcePositionAndKeyPath is set based on the provided source position tree, which contains
    the source position information for the object and its children.

    Args:
        obj (Any): The object to populate the source position and key path for.
        source_position_tree (Optional[SourcePositionTree]): The tree node containing the source
            position information for the object and its children.
        key_path (KeyPath): The path of keys that lead to the current object.
    """
    if isinstance(obj, HasSourcePositionAndKeyPath):
        check.invariant(
            obj._source_position_and_key_path is None,  # noqa: SLF001
            "Cannot call populate_source_position_and_key_paths() more than once on the same object",
        )

        if source_position_tree is not None:
            object.__setattr__(
                obj,
                "_source_position_and_key_path",
                SourcePositionAndKeyPath(key_path, source_position_tree.position),
            )

    if source_position_tree is None:
        return

    for child_key_segment, child_tree in source_position_tree.children.items():
        try:
            child_obj = cast("Any", obj)[child_key_segment]
        except TypeError:
            if not isinstance(child_key_segment, str):
                raise
            child_obj = getattr(obj, child_key_segment)

        populate_source_position_and_key_paths(
            child_obj, child_tree, [*key_path, child_key_segment]
        )



import functools
import glob
from collections.abc import Mapping, Sequence
from typing import Any, Optional, cast

import yaml

import dagster_shared.check as check
from dagster_shared.merger import deep_merge_dicts
from dagster_shared.yaml_utils.source_position import (
    KeyPathSegment,
    LineCol,
    SourcePosition,
    SourcePositionTree,
    ValueAndSourcePositionTree,
)

# Removing this implicit loader stops YAML-loading from returning datetime objects
YAML_TIMESTAMP_TAG = "tag:yaml.org,2002:timestamp"
YAML_STR_TAG = "tag:yaml.org,2002:str"


class _CanRemoveImplicitResolver:
    # Adds a "remove_implicit_resolver" method that can be used to selectively
    # disable default PyYAML resolvers

    @classmethod
    def remove_implicit_resolver(cls, tag):
        # See https://github.com/yaml/pyyaml/blob/master/lib/yaml/resolver.py#L26 for inspiration
        if "yaml_implicit_resolvers" not in cls.__dict__:
            implicit_resolvers = {}
            for key in cls.yaml_implicit_resolvers:
                implicit_resolvers[key] = cls.yaml_implicit_resolvers[key][:]
            cls.yaml_implicit_resolvers = implicit_resolvers

        for ch, mappings in cls.yaml_implicit_resolvers.items():
            cls.yaml_implicit_resolvers[ch] = [
                (existing_tag, regexp) for existing_tag, regexp in mappings if existing_tag != tag
            ]


# Handles strings with leading 0s being unexpectedly parsed as octal ints
# See: https://github.com/yaml/pyyaml/issues/98#issuecomment-436814271
def _octal_string_representer(dumper, value):
    if value.startswith("0"):
        return dumper.represent_scalar(YAML_STR_TAG, value, style="'")
    return dumper.represent_scalar(YAML_STR_TAG, value)


class DagsterRunConfigYamlLoader(yaml.SafeLoader, _CanRemoveImplicitResolver):
    pass


DagsterRunConfigYamlLoader.remove_implicit_resolver(YAML_TIMESTAMP_TAG)


class DagsterRunConfigYamlDumper(yaml.SafeDumper, _CanRemoveImplicitResolver):
    pass


DagsterRunConfigYamlDumper.remove_implicit_resolver(YAML_TIMESTAMP_TAG)
DagsterRunConfigYamlDumper.add_representer(str, _octal_string_representer)


def load_yaml_from_globs(
    *globs: str, loader: type[yaml.SafeLoader] = DagsterRunConfigYamlLoader
) -> Mapping[object, object]:
    return load_yaml_from_glob_list(list(globs), loader=loader)


def load_yaml_from_glob_list(
    glob_list: Sequence[str], loader: type[yaml.SafeLoader] = DagsterRunConfigYamlLoader
) -> Mapping[object, object]:
    check.sequence_param(glob_list, "glob_list", of_type=str)

    all_files_list: list[str] = []

    for env_file_pattern in glob_list:
        all_files_list.extend(glob.glob(env_file_pattern))

    return merge_yamls(all_files_list, loader=loader)


def merge_yamls(
    file_list: Sequence[str], loader: type[yaml.SafeLoader] = DagsterRunConfigYamlLoader
) -> dict[object, object]:
    """Combine a list of YAML files into a dictionary.

    Args:
        file_list (List[str]): List of YAML filenames

    Returns:
        dict: Merged dictionary from combined YAMLs

    Raises:
        yaml.YAMLError: When one of the YAML documents is invalid and has a parse error.
    """
    check.sequence_param(file_list, "file_list", of_type=str)

    merged: dict[object, object] = {}

    for yaml_file in file_list:
        yaml_dict = load_yaml_from_path(yaml_file, loader=loader) or {}

        if isinstance(yaml_dict, dict):
            merged = deep_merge_dicts(merged, yaml_dict)
        else:
            check.failed(
                f"Expected YAML from file {yaml_file} to parse to dictionary, "
                f'instead got: "{yaml_dict}"'
            )
    return merged


def merge_yaml_strings(
    yaml_strs: Sequence[str], loader: type[yaml.SafeLoader] = DagsterRunConfigYamlLoader
) -> dict[object, object]:
    """Combine a list of YAML strings into a dictionary.  Right-most overrides left-most.

    Args:
        yaml_strs (List[str]): List of YAML strings

    Returns:
        dict: Merged dictionary from combined YAMLs

    Raises:
        yaml.YAMLError: When one of the YAML documents is invalid and has a parse error.
    """
    check.sequence_param(yaml_strs, "yaml_strs", of_type=str)

    # Read YAML strings.
    yaml_dicts = list([yaml.load(y, Loader=loader) for y in yaml_strs])

    for yaml_dict in yaml_dicts:
        check.invariant(
            isinstance(yaml_dict, dict),
            f'Expected YAML dictionary, instead got: "{yaml_dict!s}"',
        )

    return functools.reduce(deep_merge_dicts, yaml_dicts, {})


def load_yaml_from_path(
    path: str, loader: type[yaml.SafeLoader] = DagsterRunConfigYamlLoader
) -> object:
    check.str_param(path, "path")
    with open(path, encoding="utf8") as ff:
        return yaml.load(ff, Loader=loader)


def load_run_config_yaml(yaml_str: str) -> Mapping[str, object]:
    return yaml.load(yaml_str, Loader=DagsterRunConfigYamlLoader)


def dump_run_config_yaml(run_config: Mapping[str, Any], sort_keys: bool = True) -> str:
    return yaml.dump(
        run_config,
        Dumper=DagsterRunConfigYamlDumper,
        default_flow_style=False,
        allow_unicode=True,
        sort_keys=sort_keys,
    )


def parse_yamls_with_source_position(
    src: str,
    filename: str = "<string>",
    implicits_to_remove: Sequence[str] = [YAML_TIMESTAMP_TAG],
) -> list[ValueAndSourcePositionTree]:
    """Parse YAML source with source position information.
    This function takes a YAML source string and an optional filename, and returns a
    `ValueAndSourcePositionTree` object. The `ValueAndSourcePositionTree` contains the
    parsed YAML value and a `SourcePositionTree` that holds the source position information for
    the YAML structure.

    The function uses a custom YAML loader (`SourcePositionLoader`) that tracks the source position
    information for each YAML node. It also allows for the removal of specific implicit resolvers,
    such as the default timestamp resolver.

    Args:
        src (str): The YAML source string to be parsed.
        filename (str): The filename associated with the YAML source, used for error reporting.
            Defaults to "<string>" if not provided.
        implicits_to_remove (list[str]): A list of YAML implicit resolvers to remove. Defaults to
            removing the timestamp resolver.

    Returns:
        ValueAndSourcePositionTree: An object containing the parsed YAML value and the corresponding
            source position information.
    """

    class SourcePositionLoader(yaml.SafeLoader, _CanRemoveImplicitResolver):
        def construct_object(self, node: Any, deep: Any = False) -> ValueAndSourcePositionTree:
            """Returns a ValueAndSourcePositionTree object where:
            - Its value is what the regular yaml.SafeLoader returns
            - Its source_position_tree is a tree where each node corresponds to a node in the YAML
              tree.
            """
            source_position = SourcePosition(
                filename=filename,
                start=LineCol(line=node.start_mark.line + 1, col=node.start_mark.column + 1),
                end=LineCol(line=node.end_mark.line + 1, col=node.end_mark.column + 1),
            )

            value = super().construct_object(node, True)
            # If value is a collection, then all of its elements will be ValueAndSourcePositionTree
            # instances. We need to do two things:
            # - Strip out the enclosing ValueAndSourcePositionTrees, so we're just returning a
            #   collection of the raw values
            # - Assemble the source position subtrees into a super-tree

            if isinstance(value, dict):
                dict_with_raw_values: dict[Any, Any] = {}
                child_trees: dict[KeyPathSegment, SourcePositionTree] = {}
                for k, v in cast(
                    "Mapping[ValueAndSourcePositionTree, ValueAndSourcePositionTree]", value
                ).items():
                    dict_with_raw_values[k.value] = v.value
                    child_trees[k.value] = SourcePositionTree(
                        position=k.source_position_tree.position,
                        children=v.source_position_tree.children,
                    )
                return ValueAndSourcePositionTree(
                    dict_with_raw_values,
                    SourcePositionTree(position=source_position, children=child_trees),
                )
            if isinstance(value, list):
                list_with_raw_values: list[Any] = []
                child_trees: dict[KeyPathSegment, SourcePositionTree] = {}
                for i, v in enumerate(cast("Sequence[ValueAndSourcePositionTree]", value)):
                    list_with_raw_values.append(v.value)
                    child_trees[i] = v.source_position_tree
                return ValueAndSourcePositionTree(
                    list_with_raw_values,
                    SourcePositionTree(position=source_position, children=child_trees),
                )

            return ValueAndSourcePositionTree(
                value,
                SourcePositionTree(position=source_position, children={}),
            )

    for implicit_to_remove in implicits_to_remove:
        SourcePositionLoader.remove_implicit_resolver(implicit_to_remove)

    try:
        value_and_tree_nodes = list(yaml.load_all(src, Loader=SourcePositionLoader))
    except yaml.MarkedYAMLError as e:
        if e.context_mark is not None:
            e.context_mark.name = filename
        if e.problem_mark is not None:
            e.problem_mark.name = filename
        raise e

    return check.is_list(value_and_tree_nodes, of_type=ValueAndSourcePositionTree)


def try_parse_yaml_with_source_position(
    src: str,
    filename: str = "<string>",
    implicits_to_remove: Sequence[str] = [YAML_TIMESTAMP_TAG],
) -> Optional[ValueAndSourcePositionTree]:
    yamls = parse_yamls_with_source_position(
        src,
        filename=filename,
        implicits_to_remove=implicits_to_remove,
    )
    return next(iter(yamls)) if yamls else None


def parse_yaml_with_source_position(
    src: str,
    filename: str = "<string>",
    implicits_to_remove: Sequence[str] = [YAML_TIMESTAMP_TAG],
) -> ValueAndSourcePositionTree:
    return check.not_none(
        try_parse_yaml_with_source_position(
            src,
            filename=filename,
            implicits_to_remove=implicits_to_remove,
        ),
        f"Not able to parse YAML with source position in filename {filename}",
    )



# mypy does not support recursive types, so "cause" has to be typed `Any`
import traceback
from collections.abc import Sequence
from typing import Any, Callable, NamedTuple, Optional

from typing_extensions import Self

from dagster_shared.serdes.serdes import whitelist_for_serdes


# TODO: Eventually we need to move the rest of the code in dagster._utils.error into dagster_shared,
# but that is a significant refactor. Currently we're just moving SerializableErrorInfo so that we
# can use it with `dagster-dg`.
@whitelist_for_serdes
class SerializableErrorInfo(
    NamedTuple(
        "SerializableErrorInfo",
        [
            ("message", str),
            ("stack", Sequence[str]),
            ("cls_name", Optional[str]),
            ("cause", Any),
            ("context", Any),
        ],
    )
):
    # serdes log
    # * added cause - default to None in constructor to allow loading old entries
    # * added context - default to None for similar reasons
    #
    def __new__(
        cls,
        message: str,
        stack: Sequence[str],
        cls_name: Optional[str],
        cause: Optional["SerializableErrorInfo"] = None,
        context: Optional["SerializableErrorInfo"] = None,
    ):
        return super().__new__(cls, message, stack, cls_name, cause, context)

    def __str__(self) -> str:
        return self.to_string()

    def to_string(self) -> str:
        stack_str = "\nStack Trace:\n" + "".join(self.stack) if self.stack else ""
        cause_str = (
            "\nThe above exception was caused by the following exception:\n"
            + self.cause.to_string()
            if self.cause
            else ""
        )
        context_str = (
            "\nThe above exception occurred during handling of the following exception:\n"
            + self.context.to_string()
            if self.context
            else ""
        )

        return f"{self.message}{stack_str}{cause_str}{context_str}"

    def to_exception_message_only(self) -> "SerializableErrorInfo":
        """Return a new SerializableErrorInfo with only the message and cause set.

        This is done in cases when the context about the error should not be exposed to the user.
        """
        return SerializableErrorInfo(message=self.message, stack=[], cls_name=self.cls_name)

    @classmethod
    def from_traceback(cls, tb: traceback.TracebackException) -> Self:
        return cls(
            # usually one entry, multiple lines for SyntaxError
            message="".join(list(tb.format_exception_only())),
            stack=tb.stack.format(),
            cls_name=tb.exc_type.__name__ if tb.exc_type is not None else None,
            cause=cls.from_traceback(tb.__cause__) if tb.__cause__ else None,
            context=cls.from_traceback(tb.__context__)
            if tb.__context__ and not tb.__suppress_context__
            else None,
        )


DAGSTER_FRAMEWORK_SUBSTRINGS = [
    "/site-packages/dagster",
    "/python_modules/dagster",
    "/python_modules/libraries/dagster",
]

IMPORT_MACHINERY_SUBSTRINGS = [
    "importlib/__init__.py",
    "importlib/metadata/__init__.py",
    "importlib._bootstrap",
]


NO_HINT = lambda _, __: None


def remove_system_frames_from_error(
    error_info: SerializableErrorInfo,
    build_system_frame_removed_hint: Callable[[bool, int], Optional[str]] = NO_HINT,
) -> SerializableErrorInfo:
    """Remove system frames from a SerializableErrorInfo, including Dagster framework boilerplate
    and import machinery, which are generally not useful for users to debug their code.
    """
    return remove_matching_lines_from_error_info(
        error_info,
        DAGSTER_FRAMEWORK_SUBSTRINGS + IMPORT_MACHINERY_SUBSTRINGS,
        build_system_frame_removed_hint,
    )


def make_simple_frames_removed_hint(
    additional_first_hint_warning: Optional[str] = None,
) -> Callable[[bool, int], Optional[str]]:
    def frames_removed_hint(is_first_hidden_frame: bool, num_hidden_frames: int) -> Optional[str]:
        base_hint = f"{num_hidden_frames} dagster system frames hidden"
        if is_first_hidden_frame and additional_first_hint_warning:
            return f"  [{base_hint}, {additional_first_hint_warning}]\n"
        else:
            return f"  [{base_hint}]\n"

    return frames_removed_hint


def remove_matching_lines_from_error_info(
    error_info: SerializableErrorInfo,
    match_substrs: Sequence[str],
    build_system_frame_removed_hint: Callable[[bool, int], Optional[str]],
) -> SerializableErrorInfo:
    """Utility which truncates a stacktrace to drop lines which match the given strings.
    This is useful for e.g. removing Dagster framework lines from a stacktrace that
    involves user code.

    Args:
        error_info (SerializableErrorInfo): The error info to truncate
        matching_lines (Sequence[str]): The lines to truncate from the stacktrace

    Returns:
        SerializableErrorInfo: A new error info with the stacktrace truncated
    """
    return error_info._replace(
        stack=remove_matching_lines_from_stack_trace(
            error_info.stack, match_substrs, build_system_frame_removed_hint
        ),
        cause=(
            remove_matching_lines_from_error_info(
                error_info.cause, match_substrs, build_system_frame_removed_hint
            )
            if error_info.cause
            else None
        ),
        context=(
            remove_matching_lines_from_error_info(
                error_info.context, match_substrs, build_system_frame_removed_hint
            )
            if error_info.context
            else None
        ),
    )


def remove_matching_lines_from_stack_trace(
    stack: Sequence[str],
    matching_lines: Sequence[str],
    build_system_frame_removed_hint: Callable[[bool, int], Optional[str]],
) -> Sequence[str]:
    ctr = 0
    out = []
    is_first_hidden_frame = True

    for i in range(len(stack)):
        if not _line_contains_matching_string(stack[i], matching_lines):
            if ctr > 0:
                hint = build_system_frame_removed_hint(is_first_hidden_frame, ctr)
                is_first_hidden_frame = False
                if hint:
                    out.append(hint)
            ctr = 0
            out.append(stack[i])
        else:
            ctr += 1

    if ctr > 0:
        hint = build_system_frame_removed_hint(is_first_hidden_frame, ctr)
        if hint:
            out.append(hint)

    return out


def _line_contains_matching_string(line: str, matching_strings: Sequence[str]):
    split_by_comma = line.split(",")
    if not split_by_comma:
        return False

    file_portion = split_by_comma[0]
    return any(framework_substring in file_portion for framework_substring in matching_strings)



import contextlib
import os
import random
import socket
from collections.abc import Iterator, Mapping
from typing import TypeVar

T = TypeVar("T")


def _find_free_port_in_range(start: int, end: int) -> int:
    ports_to_try = list(range(start, end + 1))
    with contextlib.closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
        random.shuffle(ports_to_try)
        for port in ports_to_try:
            try:
                s.bind(("", port))
                s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
                return port
            except OSError:
                continue

        raise Exception(f"No free ports found in range {start}-{end}")


def find_free_port() -> int:
    port_range = os.getenv("DAGSTER_PORT_RANGE")
    if port_range:
        split_range = port_range.split("-")
        if len(split_range) != 2:
            raise Exception("DAGSTER_PORT_RANGE must be of the form 'start-end'")
        return _find_free_port_in_range(int(split_range[0]), int(split_range[1]))

    with contextlib.closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
        s.bind(("", 0))
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        return s.getsockname()[1]


def remove_none_recursively(obj: T) -> T:
    """Remove none values from a dict. This can be used to support comparing provided config vs.
    config we retrive from kubernetes, which returns all fields, even those which have no value
    configured.
    """
    if isinstance(obj, (list, tuple, set)):
        return type(obj)(remove_none_recursively(x) for x in obj if x is not None)
    elif isinstance(obj, dict):
        return type(obj)(
            (remove_none_recursively(k), remove_none_recursively(v))
            for k, v in obj.items()
            if k is not None and v is not None
        )
    else:
        return obj


@contextlib.contextmanager
def environ(env: Mapping[str, str]) -> Iterator[None]:
    """Temporarily set environment variables inside the context manager and
    fully restore previous environment afterwards.
    """
    previous_values = {key: os.getenv(key) for key in env}
    for key, value in env.items():
        if value is None:
            if key in os.environ:
                del os.environ[key]
        else:
            os.environ[key] = value
    try:
        yield
    finally:
        for key, value in previous_values.items():
            if value is None:
                if key in os.environ:
                    del os.environ[key]
            else:
                os.environ[key] = value


def get_boolean_string_value(tag_value: str):
    return tag_value.lower() not in {"false", "none", "0", ""}



from collections.abc import Sequence
from typing import TYPE_CHECKING, Any, Literal, Optional, Union

from pydantic import (
    BaseModel,
    TypeAdapter,
    ValidationError,
    model_validator as model_validator,
)
from pydantic_core import PydanticUndefined as _PydanticUndefined

PydanticUndefined = _PydanticUndefined


if TYPE_CHECKING:
    from pydantic.fields import ModelField  # type: ignore


class ModelFieldCompat:
    """Wraps a Pydantic model field to provide a consistent interface for accessing
    metadata and annotations between Pydantic 1 and 2.
    """

    def __init__(self, field) -> None:
        self.field: ModelField = field

    @property
    def annotation(self) -> type:
        return self.field.annotation

    @property
    def metadata(self) -> list[str]:
        return getattr(self.field, "metadata", [])

    @property
    def alias(self) -> Optional[str]:
        return self.field.alias

    @property
    def serialization_alias(self) -> Optional[str]:
        return getattr(self.field, "serialization_alias", None)

    @property
    def validation_alias(self) -> Optional[str]:
        return getattr(self.field, "validation_alias", None)

    @property
    def default(self) -> Any:
        return self.field.default

    @property
    def description(self) -> Optional[str]:
        return getattr(self.field, "description", None)

    def is_required(self) -> bool:
        return self.field.is_required()

    @property
    def discriminator(self) -> Optional[str]:
        if hasattr(self.field, "discriminator"):
            return self.field.discriminator if hasattr(self.field, "discriminator") else None


def model_fields(model: type[BaseModel]) -> dict[str, ModelFieldCompat]:
    """Returns a dictionary of fields for a given pydantic model, wrapped
    in a compat class to provide a consistent interface between Pydantic 1 and 2.
    """
    fields = getattr(model, "model_fields", None)
    if not fields:
        fields = getattr(model, "__fields__")

    return {k: ModelFieldCompat(v) for k, v in fields.items()}


def model_config(model: type[BaseModel]):
    """Returns the config for a given pydantic model, wrapped such that it has
    a Pydantic 2-style interface for accessing config values.
    """
    return getattr(model, "model_config")


def build_validation_error(
    base_error: ValidationError,
    line_errors: list,
    hide_input: bool,
    input_type: Literal["python", "json"],
) -> ValidationError:
    return ValidationError.from_exception_data(
        title=base_error.title,
        line_errors=line_errors,
        input_type=input_type,
        hide_input=hide_input,
    )


def json_schema_from_type(model_type: Union[type[BaseModel], type[Sequence[BaseModel]]]):
    """Pydantic version stable way to get the JSON schema for a Pydantic model."""
    return TypeAdapter(model_type).json_schema()



import inspect
import os
from abc import ABC
from collections import namedtuple
from collections.abc import Mapping
from functools import partial
from typing import (  # noqa: UP035
    TYPE_CHECKING,
    Any,
    Callable,
    Generic,
    Iterator,
    NamedTuple,
    Optional,
    Tuple,  # noqa: F401
    Type,  # noqa: F401
    TypeVar,
    Union,
    overload,
)

from typing_extensions import Self, dataclass_transform

import dagster_shared.check as check
from dagster_shared.check.builder import (
    INJECTED_DEFAULT_VALS_LOCAL_VAR,
    EvalContext,
    build_args_and_assignment_strs,
    build_check_call_str,
)
from dagster_shared.check.record import RECORD_MARKER_FIELD, RECORD_MARKER_VALUE, is_record

ImportFrom = check.ImportFrom  # re-expose for convenience
TType = TypeVar("TType", bound=type)
TVal = TypeVar("TVal")


_RECORD_ANNOTATIONS_FIELD = "__record_annotations__"
_RECORD_DEFAULTS_FIELD = "__record_defaults__"
_CHECKED_NEW = "__checked_new__"
_DEFAULTS_NEW = "__defaults_new__"
_NAMED_TUPLE_BASE_NEW_FIELD = "__nt_new__"
_REMAPPING_FIELD = "__field_remap__"
_ORIGINAL_CLASS_FIELD = "__original_class__"
_KW_ONLY_FIELD = "__kw_only__"


_sample_nt = namedtuple("_canary", "x")
# use a sample to avoid direct private imports (_collections._tuplegetter)
_tuple_getter_type = type(getattr(_sample_nt, "x"))


def _get_field_set_and_defaults(
    cls: type,
    kw_only: bool,
) -> tuple[Mapping[str, Any], Mapping[str, Any]]:
    field_set = getattr(cls, "__annotations__", {})
    defaults = {}

    last_defaulted_field = None
    for name in field_set.keys():
        if hasattr(cls, name):
            attr_val = getattr(cls, name)
            if isinstance(attr_val, property):
                check.invariant(
                    attr_val.__isabstractmethod__,
                    f"Conflicting non-abstract @property for field {name} on record {cls.__name__}."
                    "Add the the @abstractmethod decorator to make it abstract.",
                )
            # When doing record inheritance, filter out tuplegetters from parents.
            # This workaround only seems needed for py3.9
            elif not isinstance(attr_val, _tuple_getter_type):
                check.invariant(
                    not inspect.isfunction(attr_val),
                    f"Conflicting function for field {name} on record {cls.__name__}. "
                    "If you are trying to set a function as a default value "
                    "you will have to override __new__.",
                )
                defaults[name] = attr_val
                last_defaulted_field = name
                continue

        # fall through here means no default set
        if last_defaulted_field and not kw_only:
            check.failed(
                "Fields without defaults cannot appear after fields with default values. "
                f"Field {name} has no default after {last_defaulted_field} with default value."
            )

    for base in cls.__bases__:
        if is_record(base):
            original_base = getattr(base, _ORIGINAL_CLASS_FIELD)
            base_kw_only = getattr(base, _KW_ONLY_FIELD)
            check.invariant(
                kw_only == base_kw_only,
                "Can not inherit from a parent @record with different kw_only setting.",
            )
            base_field_set, base_defaults = _get_field_set_and_defaults(original_base, kw_only)
            field_set = {**base_field_set, **field_set}
            defaults = {**base_defaults, **defaults}

    return field_set, defaults


def _namedtuple_record_transform(
    cls: TType,
    *,
    checked: bool,
    with_new: bool,
    decorator_frames: int,
    field_to_new_mapping: Optional[Mapping[str, str]],
    kw_only: bool,
) -> TType:
    """Transforms the input class in to one that inherits a generated NamedTuple base class
    and:
        * bans tuple methods that don't make sense for a record object
        * creates a run time checked __new__  (optional).
    """
    field_set, defaults = _get_field_set_and_defaults(cls, kw_only)

    base = NamedTuple(f"_{cls.__name__}", field_set.items())
    nt_new = base.__new__

    generated_new = None
    if checked:
        eval_ctx = EvalContext.capture_from_frame(
            1 + decorator_frames,
            # inject default values in to the local namespace for reference in generated __new__
            add_to_local_ns={INJECTED_DEFAULT_VALS_LOCAL_VAR: defaults},
        )
        generated_new = JitCheckedNew(
            field_set,
            defaults,
            eval_ctx,
            new_frames=1 if with_new else 0,
            kw_only=kw_only,
        )
    elif defaults:
        # allow arbitrary ordering of default values by generating a kwarg only __new__ impl
        eval_ctx = EvalContext(
            global_ns={},
            # inject default values in to the local namespace for reference in generated __new__
            local_ns={INJECTED_DEFAULT_VALS_LOCAL_VAR: defaults},
            lazy_imports={},
        )
        generated_new = eval_ctx.compile_fn(
            _build_defaults_new(field_set, defaults, kw_only),
            _DEFAULTS_NEW,
        )

    # the default namedtuple record cannot handle subclasses that have different fields from their
    # parents if both are records
    base.__repr__ = _repr
    nt_iter = base.__iter__

    # disable iteration unless the debugger is running which uses iteration to display values
    if not os.getenv("DEBUGPY_RUNNING"):
        base.__iter__ = _banned_iter
        base.__getitem__ = _banned_idx

    # these will override an implementation on the class if it exists
    new_class_dict = {
        **{n: getattr(base, n) for n in field_set.keys()},
        "_fields": base._fields,
        "__hidden_iter__": nt_iter,
        "__hidden_replace__": base._replace,
        RECORD_MARKER_FIELD: RECORD_MARKER_VALUE,
        _RECORD_ANNOTATIONS_FIELD: field_set,
        _RECORD_DEFAULTS_FIELD: defaults,
        _NAMED_TUPLE_BASE_NEW_FIELD: nt_new,
        _REMAPPING_FIELD: field_to_new_mapping or {},
        _ORIGINAL_CLASS_FIELD: cls,
        _KW_ONLY_FIELD: kw_only,
        "__reduce__": _reduce,
        # functools doesn't work, so manually update_wrapper
        "__module__": cls.__module__,
        "__qualname__": cls.__qualname__,
        "__annotations__": field_set,
        "__doc__": cls.__doc__,
        "__get_pydantic_core_schema__": _pydantic_core_schema,
    }

    # Due to MRO issues, we can not support both @record_custom __new__ and record inheritance
    # so enforce these rules and place the generated new in the appropriate place in the class hierarchy.
    if with_new:
        if not _defines_own_new(cls):
            # verify the alignment since it impacts frame capture
            check.failed(
                f"Expected __new__ on {cls.__name__}, add it or switch from the @record_custom decorator to @record."
            )
        if is_record(cls):
            parent = next(c for c in cls.__mro__ if is_record(c))
            check.failed(
                f"@record_custom can not be used with @record inheritance. {cls.__name__} is a child of @record {parent.__name__}."
            )

        # For records with custom new, put the generated new on the NT base class
        if generated_new:
            base.__new__ = generated_new

    elif generated_new:
        if _defines_own_new(cls):
            check.failed(
                f"Found a custom __new__ on @record {cls}, use @record_custom with IHaveNew instead."
            )
        for c in cls.__mro__:
            if c is cls:
                continue

            if is_record(c) and _defines_own_new(c):
                check.failed(
                    "@record can not inherit from @record_custom. "
                    f"@record {cls.__name__} inherits from @record_custom {c.__name__}"
                )

        # For regular @records, which may inherit from other records, put new on the generated class to avoid
        # MRO resolving to the wrong NT base
        new_class_dict["__new__"] = generated_new
        new_class_dict["_make"] = base._make

    # make it so instances of records with no fields still evaluate to true
    if len(field_set) < 1:
        new_class_dict["__bool__"] = _true

    new_type = type(
        cls.__name__,
        (cls, base),
        new_class_dict,
    )

    # setting this in the dict above does not work for some reason,
    # so set it directly after instantiation
    if hasattr(cls, "__parameters__"):
        setattr(new_type, "__parameters__", cls.__parameters__)

    return new_type  # type: ignore


@overload
def record(
    cls: TType,
) -> TType: ...  # Overload for using decorator with no ().


@overload
def record(
    *,
    checked: bool = True,
    kw_only: bool = True,
) -> Callable[[TType], TType]: ...  # Overload for using decorator used with args.


@dataclass_transform(
    kw_only_default=True,
    frozen_default=True,
)
def record(
    cls: Optional[TType] = None,
    *,
    checked: bool = True,
    kw_only: bool = True,
) -> Union[TType, Callable[[TType], TType]]:
    """A class decorator that will create an immutable record class based on the defined fields.

    Args:
        checked: Whether or not to generate runtime type checked construction (default True).
        kw_only: Whether or not the generated __new__ is kwargs only (default True).
    """
    if cls:
        return _namedtuple_record_transform(
            cls,
            checked=checked,
            with_new=False,
            decorator_frames=1,
            field_to_new_mapping=None,
            kw_only=kw_only,
        )
    else:
        return partial(
            _namedtuple_record_transform,
            checked=checked,
            with_new=False,
            decorator_frames=0,
            field_to_new_mapping=None,
            kw_only=kw_only,
        )


@overload
def record_custom(
    cls: TType,
) -> TType: ...  # Overload for using decorator with no ().


@overload
def record_custom(
    *,
    checked: bool = True,
    field_to_new_mapping: Optional[Mapping[str, str]] = None,
) -> Callable[[TType], TType]: ...  # Overload for using decorator used with args.


def record_custom(
    cls: Optional[TType] = None,
    *,
    checked: bool = True,
    field_to_new_mapping: Optional[Mapping[str, str]] = None,
) -> Union[TType, Callable[[TType], TType]]:
    """Variant of the record decorator to use to opt out of the dataclass_transform decorator behavior.
    This is done when overriding __new__ so that the type checker knows that is what is used.

    @record_custom
    class Coerced(IHaveNew):
        name: str

        def __new__(cls, name: Optional[str] = None)
            if not name:
                name = "bob"

            return super().__new__(
                cls,
                name=name,
            )



    It would have been cool if we could do that with an argument and @overload but
    from https://peps.python.org/pep-0681/ "When applied to an overload,
    the dataclass_transform decorator still impacts all usage of the function."
    """
    if cls:
        return _namedtuple_record_transform(
            cls,
            checked=checked,
            with_new=True,
            decorator_frames=1,
            field_to_new_mapping=field_to_new_mapping,
            kw_only=True,
        )
    else:
        return partial(
            _namedtuple_record_transform,
            checked=checked,
            with_new=True,
            decorator_frames=0,
            field_to_new_mapping=field_to_new_mapping,
            kw_only=True,
        )


class IHaveNew:
    """Marker class to be used when overriding new in @record_custom classes to prevent
    type errors when calling super().__new__.
    """

    if TYPE_CHECKING:

        def __new__(cls, **kwargs) -> Self: ...

        # let type checker know these objects are sortable (by way of being a namedtuple)
        def __lt__(self, other) -> bool: ...


def has_generated_new(obj) -> bool:
    return obj.__new__.__name__ in (_DEFAULTS_NEW, _CHECKED_NEW)


def get_record_annotations(obj) -> Mapping[str, type]:
    check.invariant(is_record(obj), "Only works for @record decorated classes")
    return getattr(obj, _RECORD_ANNOTATIONS_FIELD)


def get_record_defaults(obj) -> Mapping[str, Any]:
    check.invariant(is_record(obj), "Only works for @record decorated classes")
    return getattr(obj, _RECORD_DEFAULTS_FIELD)


def get_original_class(obj):
    check.invariant(is_record(obj), "Only works for @record decorated classes")
    return getattr(obj, _ORIGINAL_CLASS_FIELD)


def as_dict(obj) -> Mapping[str, Any]:
    """Creates a dict representation of the record based on the fields."""
    check.invariant(is_record(obj), "Only works for @record decorated classes")

    return {key: value for key, value in zip(obj._fields, obj.__hidden_iter__())}


def as_dict_for_new(obj) -> Mapping[str, Any]:
    """Creates a dict representation of the record with field_to_new_mapping applied."""
    check.invariant(is_record(obj), "Only works for @record decorated classes")

    remap = getattr(obj, _REMAPPING_FIELD)
    from_obj = {}
    for k, v in as_dict(obj).items():
        if k in remap:
            from_obj[remap[k]] = v
        else:
            from_obj[k] = v

    return from_obj


def copy(record: TVal, **kwargs) -> TVal:
    """Create a new instance of this record type using its constructor with
    the original records values overrode with new values specified as key args.
    """
    return record.__class__(
        **{
            **as_dict_for_new(record),
            **kwargs,
        }
    )


def replace(obj: TVal, **kwargs) -> TVal:
    """Create a new instance of this record type using the record constructor directly,
    (bypassing any custom __new__ impl) with the original records values overrode with
    new values specified by keyword args.

    This emulates the behavior of namedtuple _replace.
    """
    check.invariant(is_record(obj))
    cls = obj.__class__

    # if we have runtime type checking, go through that to vet new field values
    if hasattr(cls, _CHECKED_NEW):
        target = _CHECKED_NEW
    else:
        target = _NAMED_TUPLE_BASE_NEW_FIELD

    return getattr(cls, target)(
        obj.__class__,
        **{**as_dict(obj), **kwargs},
    )


class LegacyNamedTupleMixin(ABC):
    """Mixin to ease migration by adding NamedTuple utility methods.
    Inherit when converting an existing NamedTuple that has callsites to _replace / _asdict, ie.

    @record
    def AssetSubset(LegacyNamedTupleMixin):
        asset_key: AssetKey
        value: Union[bool, PartitionsSubset]
    """

    def _replace(self, **kwargs):
        return replace(self, **kwargs)

    def _asdict(self):
        return as_dict(self)


class JitCheckedNew:
    """Object that allows us to just-in-time compile a checked __new__ implementation on first use.
    This has two benefits:
        1. Defer processing ForwardRefs until their definitions are in scope.
        2. Avoid up-front cost for unused objects.
    """

    __name__ = _CHECKED_NEW

    def __init__(
        self,
        field_set: Mapping[str, type],
        defaults: Mapping[str, Any],
        eval_ctx: EvalContext,
        new_frames: int,
        kw_only: bool,
    ):
        self._field_set = field_set
        self._defaults = defaults
        self._eval_ctx = eval_ctx
        self._new_frames = new_frames  # how many frames of __new__ there are
        self._compiled = False
        self._kw_only = kw_only

    def __call__(self, cls, *args, **kwargs):
        if _do_defensive_checks():
            # this condition can happen during races in threaded envs so only
            # invariant when opted-in
            check.invariant(
                self._compiled is False,
                "failed to set compiled __new__ appropriately",
            )

        # update the context with callsite locals/globals to resolve
        # ForwardRefs that were unavailable at definition time.
        self._eval_ctx.update_from_frame(1 + self._new_frames)

        # ensure check is in scope
        if "check" not in self._eval_ctx.global_ns:
            self._eval_ctx.global_ns["check"] = check

        # we are double-memoizing this to handle some confusing mro issues
        # in which the _nt_base's __new__ method is not on the critical
        # path, causing this to get invoked multiple times
        compiled_fn = self._eval_ctx.compile_fn(
            self._build_checked_new_str(),
            _CHECKED_NEW,
        )
        self._compiled = True

        # replace this holder object with the compiled fn by finding where it was in the hierarchy
        for c in cls.__mro__:
            if c.__new__ is self:
                c.__new__ = compiled_fn

        return compiled_fn(cls, *args, **kwargs)

    def _build_checked_new_str(self) -> str:
        args_str, set_calls_str = build_args_and_assignment_strs(
            self._field_set.keys(),
            self._defaults,
            self._kw_only,
        )
        check_calls = []
        for name, ttype in self._field_set.items():
            call_str = build_check_call_str(
                ttype=ttype,
                name=name,
                eval_ctx=self._eval_ctx,
            )
            check_calls.append(f"{name}={call_str}")

        check_call_block = ",\n        ".join(check_calls)

        lazy_imports_str = "\n    ".join(
            f"from {module} import {t}" for t, module in self._eval_ctx.lazy_imports.items()
        )

        checked_new_str = f"""
def __checked_new__(cls{args_str}):
    {lazy_imports_str}
    {set_calls_str}
    return cls.{_NAMED_TUPLE_BASE_NEW_FIELD}(
        cls,
        {check_call_block}
    )
"""
        return checked_new_str


def _build_defaults_new(
    field_set: Mapping[str, type],
    defaults: Mapping[str, Any],
    kw_only: bool,
) -> str:
    """Build a __new__ implementation that handles default values."""
    kw_args_str, set_calls_str = build_args_and_assignment_strs(
        field_set,
        defaults,
        kw_only,
    )
    assign_str = ",\n        ".join([f"{name}={name}" for name in field_set.keys()])
    return f"""
def __defaults_new__(cls{kw_args_str}):
    {set_calls_str}
    return cls.{_NAMED_TUPLE_BASE_NEW_FIELD}(
        cls,
        {assign_str}
    )
    """


def _banned_iter(*args, **kwargs):
    raise Exception("Iteration is not allowed on `@record`s.")


def _banned_idx(*args, **kwargs):
    raise Exception("Index access is not allowed on `@record`s.")


def _true(_):
    return True


def _from_reduce(cls, kwargs):
    return cls(**kwargs)


def _reduce(self):
    # pickle support
    return _from_reduce, (self.__class__, as_dict_for_new(self))


def _repr(self) -> str:
    # the __repr__ method generated for namedtuples cannot handle subclasses that have different
    # sets of fields
    field_set = getattr(self, _RECORD_ANNOTATIONS_FIELD)
    values = [f"{field_name}={getattr(self, field_name)!r}" for field_name in field_set]
    return f"{self.__class__.__name__}({', '.join(values)})"


def _defines_own_new(cls) -> bool:
    qualname = getattr(cls.__new__, "__qualname__", None)
    if not qualname:
        return False

    qualname_parts = cls.__new__.__qualname__.split(".")
    if len(qualname_parts) < 2:
        return False

    return qualname_parts[-2] == cls.__name__


def _do_defensive_checks():
    return bool(os.getenv("DAGSTER_RECORD_DEFENSIVE_CHECKS"))


@classmethod
def _pydantic_core_schema(cls, source: Any, handler):
    """Forces pydantic_core to treat records as normal types instead of namedtuples. In particular,
    pydantic assumes that all namedtuples can be constructed with posargs, while records are kw-only.
    """
    from pydantic_core import core_schema

    return core_schema.is_instance_schema(cls)


TRecord = TypeVar("TRecord")


class NamedTupleAdapter(Generic[TRecord]):
    def _asdict(self: TRecord) -> Mapping[str, Any]:
        return as_dict(self)

    def _replace(self: TRecord, **kwargs) -> TRecord:
        return replace(self, **kwargs)

    def __iter__(self) -> Iterator:
        return tuple.__iter__(self)  # type: ignore



import json
import socket
import warnings
from collections.abc import Mapping
from typing import Any

from packaging.version import Version

from dagster_shared import check
from dagster_shared.version import __version__


class DagsterLibraryRegistry:
    _libraries: dict[str, str] = {"dagster-shared": __version__}

    @classmethod
    def register(cls, name: str, version: str, *, is_dagster_package: bool = True):
        if is_dagster_package:
            check_dagster_package_version(name, version)

        cls._libraries[name] = version

    @classmethod
    def get(cls) -> Mapping[str, str]:
        return cls._libraries.copy()


def parse_package_version(version_str: str) -> Version:
    parsed_version = Version(version_str)
    assert isinstance(parsed_version, Version)
    return parsed_version


def increment_micro_version(v: Version, interval: int) -> Version:
    major, minor, micro = v.major, v.minor, v.micro
    new_micro = micro + interval
    if new_micro < 0:
        raise ValueError(f"Micro version cannot be negative: {new_micro}")
    return Version(f"{major}.{minor}.{new_micro}")


def check_dagster_package_version(library_name: str, library_version: str) -> None:
    # This import must be internal in order for this function to be testable
    from dagster_shared.version import __version__

    parsed_lib_version = parse_package_version(library_version)
    if parsed_lib_version.release[0] < 1:
        if library_version != __version__:
            message = (
                f"Found version mismatch between `dagster-shared` ({__version__}) "
                f"and `{library_name}` ({library_version})"
            )
            warnings.warn(message)
    else:
        expected_version = core_version_from_library_version(__version__)
        if library_version != expected_version:
            message = (
                f"Found version mismatch between `dagster-shared` ({__version__}) "
                f"expected library version ({expected_version}) "
                f"and `{library_name}` ({library_version})."
            )
            warnings.warn(message)


# Use this to get the "library version" (pre-1.0 version) from the "core version" (post 1.0
# version). 16 is from the 0.16.0 that library versions stayed on when core went to 1.0.0.
def library_version_from_core_version(core_version: str) -> str:
    parsed_version = parse_package_version(core_version)

    release = parsed_version.release
    if release[0] >= 1:
        library_version = ".".join(["0", str(16 + release[1]), str(release[2])])

        if parsed_version.is_prerelease:
            library_version = library_version + "".join(
                [str(pre) for pre in check.not_none(parsed_version.pre)]
            )

        if parsed_version.is_postrelease:
            library_version = library_version + "post" + str(parsed_version.post)

        return library_version
    else:
        return core_version


def core_version_from_library_version(library_version: str) -> str:
    parsed_version = parse_package_version(library_version)

    release = parsed_version.release
    if release[0] < 1 and len(release) > 1:
        core_version = ".".join(["1", str(release[1] - 16), str(release[2])])

        if parsed_version.is_prerelease:
            core_version = core_version + "".join(
                [str(pre) for pre in check.not_none(parsed_version.pre)]
            )

        if parsed_version.is_postrelease:
            core_version = core_version + "post" + str(parsed_version.post)

        return core_version
    else:
        return library_version


class DagsterPyPiAccessError(Exception):
    pass


def get_pypi_package_data(pkg_name: str, timeout: float = 5.0) -> dict[str, Any]:
    # defer for import performance
    import urllib.request
    from urllib.error import HTTPError, URLError

    url = f"https://pypi.org/pypi/{pkg_name}/json"

    try:
        with urllib.request.urlopen(url, timeout=timeout) as response:
            if response.status != 200:
                raise DagsterPyPiAccessError(
                    f"Error: Received status code {response.status} for {pkg_name}"
                )
            return json.load(response)
    except (HTTPError, URLError, socket.timeout) as e:
        raise DagsterPyPiAccessError(f"Network error while checking {pkg_name}: {e}")
    except json.JSONDecodeError as e:
        raise DagsterPyPiAccessError(f"Invalid JSON response for {pkg_name}: {e}")


def get_published_pypi_versions(pkg_name: str, timeout: float = 5.0) -> list[Version]:
    package_data = get_pypi_package_data(pkg_name, timeout)
    return sorted(Version(k) for k in package_data["releases"].keys())
